import abc
import six


@six.add_metaclass(abc.ABCMeta)
class ContentTypeValidator (object):
    """
    Abstract interface for the provision of a method to provide a "valid" set of
    content types relative to the sub-class' function.
    """

    __slots__ = ()

    @abc.abstractmethod
    def valid_content_types(self):
        """
        :return: A set valid MIME types that are "valid" within the implementing
            class' context.
        :rtype: set[str]
        """

    def is_valid_element(self, data_element):
        """
        Check if the given DataElement instance reports a content type that
        matches one of the MIME types reported by ``valid_content_types``.

        :param smqtk.representation.DataElement data_element:
             Data element instance to check.

        :return: True if the given element has a valid content type as reported
            by ``valid_content_types``, and False if not.
        :rtype: bool
        """
        return data_element.content_type() in self.valid_content_types()

    def raise_valid_element(self, data_element, exception_type=ValueError,
                            message=None):
        """
        Check if the given data element matches a reported valid content type,
        raising the given exception class (``ValueError`` by default) if not.

        :param smqtk.representation.DataElement data_element:
             Data element instance to check.
        :param StandardError exception_type:
            Custom exception type to raise if the given element does not report
            as a valid content type. By default we raise a ``ValueError``.
        :param str message:
            Specific message to provide with a raise exception. By default
            we compose a generic message that also reports the given
            element's content type.

        :return: The unmodified input data element.
        :rtype: smqtk.representation.DataElement
        """
        if not self.is_valid_element(data_element):
            if message is None:
                message = "Data element does not match a content type " \
                          "reported as valid. Given: \"{}\". Valid types: {}." \
                          .format(data_element.content_type(),
                                  list(self.valid_content_types()))
            # noinspection PyCallingNonCallable
            # - Leave the handling of whether or not an exception is
            # constructable to the exception class being constructed (user
            # decision repercussion).
            raise exception_type(message)
        return data_element



import io
import json
import logging
import threading
import uuid
import zipfile

import six

from smqtk.algorithms.relevancy_index import RelevancyIndex
from smqtk.representation.descriptor_set.memory import MemoryDescriptorSet
from smqtk.utils import SmqtkObject
from smqtk.utils.configuration import from_config_dict


DFLT_REL_INDEX_CONFIG = {
    "type": "LibSvmHikRelevancyIndex",
    "LibSvmHikRelevancyIndex": {
        "descr_cache_filepath": None,
    }
}


class IqrSession (SmqtkObject):
    """
    Encapsulation of IQR Session related data structures with a centralized
    lock for multi-thread access.

    This object is compatible with the python with-statement, so when elements
    are to be used or modified, it should be within a with-block so race
    conditions do not occur across threads/sub-processes.

    """

    @property
    def _log(self):
        return logging.getLogger(
            '.'.join((self.__module__, self.__class__.__name__)) +
            "[%s]" % self.uuid
        )

    def __init__(self, pos_seed_neighbors=500,
                 rel_index_config=None, session_uid=None):
        """
        Initialize the IQR session

        This does not initialize the working set for ranking as there are no
        known positive descriptor examples at this time.

        Adjudications
        -------------
        Adjudications are carried through between initializations. This allows
        indexed material adjudicated through-out the lifetime of the session to
        stay relevant.

        :param pos_seed_neighbors: Number of neighbors to pull from the given
            ``nn_index`` for each positive exemplar when populating the working
            set, i.e. this value determines the size of the working set for
            IQR refinement. By default, we try to get 500 neighbors.

            Since there may be partial to significant overlap of near neighbors
            as a result of nn_index queries for positive exemplars, the working
            set may contain anywhere from this value's number of entries, to
            ``N*P``, where ``N`` is this value and ``P`` is the number of
            positive examples at the time of working set initialization.
        :type pos_seed_neighbors: int

        :param rel_index_config: Plugin configuration dictionary for the
            RelevancyIndex to use for ranking user adjudications. If `None` we
            default to using an in-memory libSVM based index using the histogram
            intersection metric.
        :type rel_index_config: None | dict

        :param session_uid: Optional manual specification of session UUID. By
            default this will be a string UUID as generated by
            ``uuid.uuid1()``.
        :type session_uid: str | uuid.UUID

        """
        self.uuid = session_uid or str(uuid.uuid1()).replace('-', '')
        self.lock = threading.RLock()

        self.pos_seed_neighbors = int(pos_seed_neighbors)

        # Local descriptor set for ranking, populated by a query to the
        #   nn_index instance.
        # Added external data/descriptors not added to this set.
        self.working_set = MemoryDescriptorSet()

        # Book-keeping set so we know what positive descriptors
        # UUIDs we've used to query the neighbor index with already.
        #: :type: set[collections.Hashable]
        self._wi_seeds_used = set()

        # Descriptor elements representing data from external sources.
        # These may be arbitrary descriptor elements not present in
        #   ``working_index``.
        #: :type: set[smqtk.representation.DescriptorElement]
        self.external_positive_descriptors = set()
        #: :type: set[smqtk.representation.DescriptorElement]
        self.external_negative_descriptors = set()

        # Descriptor references from ``working_set`` that have been
        #   adjudicated.
        # These should be sub-sets of the descriptors contained in the
        #   ``working_set``.
        #: :type: set[smqtk.representation.DescriptorElement]
        self.positive_descriptors = set()
        #: :type: set[smqtk.representation.DescriptorElement]
        self.negative_descriptors = set()

        # Sets of descriptor elements that were used in the last refinement
        #   to achieve the currently cached results, i.e. "contributed" to the
        #   current results state.
        # These sets are empty before the first refine after construction or a
        #   reset.
        #: :type: set[smqtk.representation.DescriptorElement]
        self.rank_contrib_pos = set()
        #: :type: set[smqtk.representation.DescriptorElement]
        self.rank_contrib_pos_ext = set()
        #: :type: set[smqtk.representation.DescriptorElement]
        self.rank_contrib_neg = set()
        #: :type: set[smqtk.representation.DescriptorElement]
        self.rank_contrib_neg_ext = set()

        # Mapping of a DescriptorElement in our relevancy search index (not the
        #   set that the nn_index uses) to the relevancy score given the
        #   recorded positive and negative adjudications.
        # This is None before any initialization or refinement occurs.
        #: :type: None | dict[smqtk.representation.DescriptorElement, float]
        self.results = None

        # Cache variables for views of refinement results.
        # All results as a list in order of relevancy score.
        #: :type: None | list[(smqtk.representation.DescriptorElement, float)]
        self._ordered_results = None
        #: Positively adjudicated descriptors in order of relevancy score.
        #: :type: None | list[(smqtk.representation.DescriptorElement, float)]
        self._ordered_pos = None
        # Negatively adjudicated descriptors in order of relevancy score.
        #: :type: None | list[(smqtk.representation.DescriptorElement, float)]
        self._ordered_neg = None
        # Non-adjudicated descriptors in our working set in order of
        # relevancy score.
        #: :type: None | list[(smqtk.representation.DescriptorElement, float)]
        self._ordered_non_adj = None

        #
        # Algorithm Instances [+Config]
        #
        # RelevancyIndex configuration and instance that is used for producing
        #   results.
        # This is only [re]constructed when initializing the session.
        if rel_index_config is None:
            rel_index_config = DFLT_REL_INDEX_CONFIG
        self.rel_index_config = rel_index_config
        # This is None until session initialization happens after pos/neg
        # exemplar data has been added.
        #: :type: None | smqtk.algorithms.relevancy_index.RelevancyIndex
        self.rel_index = None

    def __enter__(self):
        """
        :rtype: IqrSession
        """
        self.lock.acquire()
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()

    def external_descriptors(self, positive=(), negative=()):
        """
        Add positive/negative descriptors from external data.

        These descriptors may not be a part of our working set.

        TODO: Add ability to "remove" positive/negative external descriptors.
              See ``adjudicate`` method "un_..." parameters.

        :param positive: Iterable of descriptors from external sources to
            consider positive examples.
        :type positive:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param negative: Iterable of descriptors from external sources to
            consider negative examples.
        :type negative:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        positive = set(positive)
        negative = set(negative)
        with self.lock:
            self.external_positive_descriptors.update(positive)
            self.external_positive_descriptors.difference_update(negative)

            self.external_negative_descriptors.update(negative)
            self.external_negative_descriptors.difference_update(positive)

    def adjudicate(self, new_positives=(), new_negatives=(),
                   un_positives=(), un_negatives=()):
        """
        Update current state of working set positive and negative
        adjudications based on descriptor UUIDs.

        If the same descriptor element is listed in both new positives and
        negatives, they cancel each other out, causing that descriptor to not
        be included in the adjudication.

        The given iterables must be re-traversable. Otherwise the given
        descriptors will not be properly registered.

        :param new_positives: Descriptors of elements in our working set to
            now be considered to be positively relevant.
        :type new_positives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param new_negatives: Descriptors of elements in our working set to
            now be considered to be negatively relevant.
        :type new_negatives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param un_positives: Descriptors of elements in our working set to now
            be considered not positive any more.
        :type un_positives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param un_negatives: Descriptors of elements in our working set to now
            be considered not negative any more.
        :type un_negatives:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        # TODO: Assert that inputs are indeed in the working set?

        new_positives = set(new_positives)
        new_negatives = set(new_negatives)
        un_positives = set(un_positives)
        un_negatives = set(un_negatives)

        with self.lock:
            pos_before = set(self.positive_descriptors)
            self.positive_descriptors.update(new_positives)
            self.positive_descriptors.difference_update(un_positives)
            self.positive_descriptors.difference_update(new_negatives)
            pos_changed = pos_before != self.positive_descriptors
            if pos_changed:
                # Reset ordered positives cache if pos adjudications changed.
                self._ordered_pos = None

            neg_before = set(self.negative_descriptors)
            self.negative_descriptors.update(new_negatives)
            self.negative_descriptors.difference_update(un_negatives)
            self.negative_descriptors.difference_update(new_positives)
            neg_changed = neg_before != self.negative_descriptors
            if neg_changed:
                # Reset ordered negatives cache if neg adjudications changed.
                self._ordered_neg = None

            if pos_changed or neg_changed:
                # Reset non-adjudicated cache if anything changed.
                self._ordered_non_adj = None

    def update_working_set(self, nn_index):
        """
        Initialize or update our current working set using the given
        :class:`.NearestNeighborsIndex` instance given our current positively
        labeled descriptor elements.

        We only query from the index for new positive elements since the last
        update or reset.

        :param nn_index: :class:`.NearestNeighborsIndex` to query from.
        :type nn_index: smqtk.algorithms.NearestNeighborsIndex

        :raises RuntimeError: There are no positive example descriptors in this
            session to use as a basis for querying.

        """
        pos_examples = (self.external_positive_descriptors |
                        self.positive_descriptors)
        if len(pos_examples) == 0:
            raise RuntimeError("No positive descriptors to query the neighbor "
                               "index with.")

        # Not clearing working set because this step is intended to be
        # additive.
        updated = False

        # adding to working set
        self._log.info("Building working set using %d positive examples "
                       "(%d external, %d adjudicated)",
                       len(pos_examples),
                       len(self.external_positive_descriptors),
                       len(self.positive_descriptors))
        # TODO: parallel_map and reduce with merge-dict
        for p in pos_examples:
            if p.uuid() not in self._wi_seeds_used:
                self._log.debug("Querying neighbors to: %s", p)
                self.working_set.add_many_descriptors(
                    nn_index.nn(p, n=self.pos_seed_neighbors)[0]
                )
                self._wi_seeds_used.add(p.uuid())
                updated = True

        # Make new relevancy index
        if updated:
            self._log.info("Creating new relevancy index over working set.")
            #: :type: smqtk.algorithms.relevancy_index.RelevancyIndex
            self.rel_index = from_config_dict(
                self.rel_index_config, RelevancyIndex.get_impls()
            )
            self.rel_index.build_index(self.working_set.iterdescriptors())

    def refine(self):
        """ Refine current model results based on current adjudication state

        :raises RuntimeError: No working set has been initialized.
            :meth:`update_working_set` should have been called after
            adjudicating some positive examples.
        :raises RuntimeError: There are no adjudications to run on. We must
            have at least one positive adjudication.

        """
        with self.lock:
            if not self.rel_index:
                raise RuntimeError("No relevancy index yet. Must not have "
                                   "initialized session (no working set).")

            # combine pos/neg adjudications + added external data descriptors
            pos = self.positive_descriptors | self.external_positive_descriptors
            neg = self.negative_descriptors | self.external_negative_descriptors

            if not pos:
                raise RuntimeError("Did not find at least one positive "
                                   "adjudication.")

            self._log.debug("Ranking working set with %d pos and %d neg total "
                            "examples.", len(pos), len(neg))
            element_probability_map = self.rel_index.rank(pos, neg)
            self.results = element_probability_map
            # Record UIDs of elements used for relevancy ranking.
            # - shallow copy for separate container instance
            self.rank_contrib_pos = set(self.positive_descriptors)
            self.rank_contrib_pos_ext = set(self.external_positive_descriptors)
            self.rank_contrib_neg = set(self.negative_descriptors)
            self.rank_contrib_neg_ext = set(self.external_negative_descriptors)
            # Clear result view caches
            self._ordered_results = self._ordered_pos = self._ordered_neg = \
                self._ordered_non_adj = None

    def ordered_results(self):
        """
        Return a tuple of all working-set descriptor elements as tuples of
        ``(element, score)`` in order of descending relevancy score.

        If refinement has not yet occurred since session creation or the last
        reset, an empty tuple is returned.

        :rtype: None | tuple[(smqtk.representation.DescriptorElement, float)]
        """
        with self.lock:
            try:
                return list(self._ordered_results)
            except TypeError:
                # NoneType is not iterable
                # Cache did non exist.

                try:
                    result_items = six.iteritems(self.results)
                except AttributeError:
                    # NoneType missing items/iteritems attr
                    # No results to iterate over.
                    return list()

                r = self._ordered_results = sorted(
                    result_items,
                    key=lambda p: p[1], reverse=True
                )
                # Shallow copy of the list to protect against external mutation
                return list(r)

    def get_positive_adjudication_relevancy(self):
        """
        Return a list of the positively adjudicated descriptors as tuples of
        ``(element, score)`` in order of descending relevancy score.

        This does *not* include external positive adjudications, only
        positively adjudicated descriptors in the working set.

        If refinement has not yet occurred since session creation or the last
        reset, an empty list is returned.

        Cache is invalidated when:
        - A refinement occurs.
        - Positive adjudications change.

        :rtype: None | list[(smqtk.representation.DescriptorElement, float)]
        """
        with self.lock:
            try:
                return list(self._ordered_pos)
            except TypeError:
                # NoneType is not iterable
                # No cache yet.

                rank_contrib_pos = \
                    self.rank_contrib_pos | self.rank_contrib_pos_ext
                # Results already ordered, so only filter
                r = self._ordered_pos = list(
                    filter(lambda t: t[0] in rank_contrib_pos,
                           self.ordered_results())
                )
                # Shallow copy of the list to protect against external mutation
                return list(r)

    def get_negative_adjudication_relevancy(self):
        """
        Return a list of the negatively adjudicated descriptors as tuples of
        ``(element, score)`` in order of descending relevancy score.

        This does *not* include external negative adjudications, only
        negatively adjudicated descriptors in the working set.

        If refinement has not yet occurred since session creation or the last
        reset, an empty list is returned.

        Cache is invalidated when:
        - A refinement occurs.
        - Negative adjudications change.

        :rtype: None | list[(smqtk.representation.DescriptorElement, float)]
        """
        with self.lock:
            try:
                return list(self._ordered_neg)
            except TypeError:
                # NoneType is not iterable
                # No cache yet.

                rank_contrib_neg = \
                    self.rank_contrib_neg | self.rank_contrib_neg_ext
                # Results already ordered, so only filter
                r = self._ordered_neg = list(
                    filter(lambda t: t[0] in rank_contrib_neg,
                           self.ordered_results())
                )
                # Shallow copy of the list to protect against external mutation
                return list(r)

    def get_unadjudicated_relevancy(self):
        """
        Return a list of the non-adjudicated descriptor elements as tuples of
        ``(element, score)`` in order of descending relevancy score.

        If refinement has not yet occurred since session creation or the last
        reset, an empty list is returned.

        :rtype: None | list[(smqtk.representation.DescriptorElement, float)]
        """
        with self.lock:
            try:
                return list(self._ordered_non_adj)
            except TypeError:
                # NoneType is not iterable
                # No cache yet
                pos_and_neg = \
                    self.rank_contrib_pos | self.rank_contrib_pos_ext | \
                    self.rank_contrib_neg | self.rank_contrib_neg_ext

                # Results already ordered, so only filter
                r = self._ordered_non_adj = list(
                    filter(lambda t: t[0] not in pos_and_neg,
                           self.ordered_results())
                )
                # Shallow copy of the list to protect against external mutation
                return list(r)

    def reset(self):
        """ Reset the IQR Search state

        No positive adjudications, reload original feature data

        """
        with self.lock:
            self.working_set.clear()
            self._wi_seeds_used.clear()
            self.positive_descriptors.clear()
            self.negative_descriptors.clear()
            self.external_positive_descriptors.clear()
            self.external_negative_descriptors.clear()
            self.rank_contrib_pos.clear()
            self.rank_contrib_pos_ext.clear()
            self.rank_contrib_neg.clear()
            self.rank_contrib_neg_ext.clear()

            self.rel_index = None
            self.results = None
            self._ordered_results = self._ordered_pos = self._ordered_neg = \
                self._ordered_non_adj = None

    ###########################################################################
    # I/O Methods

    # I/O Constants. These should not be changed.
    STATE_ZIP_COMPRESSION = zipfile.ZIP_DEFLATED
    STATE_ZIP_FILENAME = "iqr_state.json"

    def get_state_bytes(self):
        """
        Get a byte representation of the current descriptor and adjudication
        state of this session.

        This does not encode current results or the relevancy index's state, but
        these can be reproduced with this state.

        :return: State representation bytes
        :rtype: bytes

        """
        def d_set_to_list(d_set):
            # Convert set of descriptors to list of tuples:
            #   [..., (uuid, type, vector), ...]
            return [(d.uuid(), d.type(), d.vector().tolist()) for d in d_set]

        with self:
            # Convert session descriptors into basic values.
            pos_d = d_set_to_list(self.positive_descriptors)
            neg_d = d_set_to_list(self.negative_descriptors)
            ext_pos_d = d_set_to_list(self.external_positive_descriptors)
            ext_neg_d = d_set_to_list(self.external_negative_descriptors)

        z_buffer = io.BytesIO()
        z = zipfile.ZipFile(z_buffer, 'w', self.STATE_ZIP_COMPRESSION)
        z.writestr(self.STATE_ZIP_FILENAME, json.dumps({
            'pos': pos_d,
            'neg': neg_d,
            'external_pos': ext_pos_d,
            'external_neg': ext_neg_d,
        }))
        z.close()
        return z_buffer.getvalue()

    def set_state_bytes(self, b, descriptor_factory):
        """
        Set this session's state to the given byte representation, resetting
        this session in the process.

        Bytes given must have been retrieved via a previous call to
        ``get_state_bytes`` otherwise this method will fail.

        Since this state may be completely different from the current state,
        this session is reset before applying the new state. Thus, any current
        ranking results are thrown away.

        :param b: Bytes to set this session's state to.
        :type b: bytes

        :param descriptor_factory: Descriptor element factory to use when
            generating descriptor elements from extracted data.
        :type descriptor_factory: smqtk.representation.DescriptorElementFactory

        :raises ValueError: The input bytes could not be loaded due to
            incompatibility.

        """
        z_buffer = io.BytesIO(b)
        z = zipfile.ZipFile(z_buffer, 'r', self.STATE_ZIP_COMPRESSION)
        if self.STATE_ZIP_FILENAME not in z.namelist():
            raise ValueError("Invalid bytes given, did not contain expected "
                             "zipped file name.")

        # Extract expected json file object
        state = json.loads(z.read(self.STATE_ZIP_FILENAME).decode())
        del z, z_buffer

        with self:
            self.reset()

            def load_descriptor(_uid, _type_str, vec_list):
                _e = descriptor_factory.new_descriptor(_type_str, _uid)
                if _e.has_vector():
                    assert _e.vector().tolist() == vec_list, \
                        "Found existing vector for UUID '%s' but vectors did " \
                        "not match."
                else:
                    _e.set_vector(vec_list)
                return _e

            # Read in raw descriptor data from the state, convert to descriptor
            # element, then store in our descriptor sets.
            for source, target in [(state['external_pos'],
                                    self.external_positive_descriptors),
                                   (state['external_neg'],
                                    self.external_negative_descriptors),
                                   (state['pos'], self.positive_descriptors),
                                   (state['neg'], self.negative_descriptors)]:
                for uid, type_str, vector_list in source:
                    e = load_descriptor(uid, type_str, vector_list)
                    target.add(e)



import abc
from collections import deque

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.representation import DescriptorElementFactory
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.utils import ContentTypeValidator


DFLT_DESCRIPTOR_FACTORY = DescriptorElementFactory(DescriptorMemoryElement, {})


class DescriptorGenerator (SmqtkAlgorithm, ContentTypeValidator):
    """
    Base abstract Feature Descriptor interface.
    """

    @abc.abstractmethod
    def _generate_arrays(self, data_iter):
        """
        Inner template method that defines the generation of descriptor vectors
        for a given iterable of data elements.

        Pre-conditions:
          - Data elements input to this method have been validated to be of at
            least one of this class's reported ``valid_content_types``.

        :param collections.Iterable[smqtk.representation.DataElement] data_iter:
            Iterable of data element instances to be described.

        :raises RuntimeError: Descriptor extraction failure of some kind.

        :return: Iterable of numpy arrays in parallel association with the
            input data elements.
        :rtype: collections.Iterable[numpy.ndarray]
        """

    def generate_arrays(self, data_iter):
        """
        Generate descriptor vector elements for **all** input data elements.

        Descriptor arrays yielded out will be parallel in association with
        the data elements input.

        **Selective Iteration**
        For situations when it is desired to access specific generator returns,
        like when only one data element is provided in order to get a single
        array out, it is strongly recommended to expand the returned generator
        into a sequence type first. For example, expanding out the generator's
        returns into a list (``list(g.generate_arrays([e]))[0]``) is
        recommended over just getting the "next" element of the returned
        generator (``next(g.generate_arrays([e]))``).
        Expansion into a sequence allows the generator to fully execute, which
        includes any functionality after the final ``yield`` statement in any
        of the underlying iterators.

        :param collections.Iterable[smqtk.representation.DataElement] data_iter:
            Iterable of DataElement instances to be described.

        :raises RuntimeError: Descriptor extraction failure of some kind.
        :raises ValueError: Given data element content was not of a valid type
            with respect to this descriptor generator implementation.

        :return: Iterator of result numpy.ndarray instances.
        :rtype: collections.Iterator[numpy.ndarray]
        """
        # Intermediate iterator for testing that content types are valid
        # TODO: Use an order-preserving call to parallel_map()?
        validated_data_iter = (self.raise_valid_element(d) for d in data_iter)
        return self._generate_arrays(validated_data_iter)

    def generate_elements(self, data_iter,
                          descr_factory=DFLT_DESCRIPTOR_FACTORY,
                          overwrite=False):
        """
        Generate DescriptorElement instances for the input data elements,
        generating new descriptors for those elements that need them, or
        optionally all input data elements.

        Descriptor elements yielded out will be parallel in association with
        the data elements input. Descriptor element UUIDs are inherited from
        the data element it was generated from.

        **Selective Iteration**
        For situations when it is desired to access specific generator returns,
        like when only one data element is provided in order to get a single
        element out, it is strongly recommended to expand the returned
        generator into a sequence type first.
        For example, expanding out the generator's returns into a list
        (``list(g.generate_elements([e]))[0]``) is recommended over just
        getting the "next" element of the returned generator
        (``next(g.generate_elements([e]))``).
        Expansion into a sequence allows the generator to fully execute, which
        includes any functionality after the final ``yield`` statement in any
        of the underlying iterators that may perform required clean-up.

        **Non-redundant Processing**
        Certain descriptor element implementations, as dictated by the input
        factory, may be connected to persistent storage in the background.
        Because of this, some descriptor elements may already "have" a vector
        on construction.
        This method, by default, only computes new descriptor vectors for data
        elements whose associated descriptor element does not report as already
        containing a vector.
        If the ``overwrite`` flag is True then descriptors are computed for all
        input data elements and are set to their respective descriptor elements
        regardless of existing vector storage.

        :param collections.Iterable[smqtk.representation.DataElement] data_iter:
            Iterable of DataElement instances to be described.
        :param smqtk.representation.DescriptorElementFactory descr_factory:
            DescriptorElementFactory instance to drive the generation of
            element instances with some parametrization.
        :param bool overwrite:
            By default, if a factory-produced DescriptorElement reports as
            containing a vector, we do not compute a descriptor again for the
            associated data element. If this is ``True``, however, we will
            generate descriptors for all input data elements, overwriting the
            vectors previously stored in the factory-produces descriptor
            elements.

        :raises RuntimeError: Descriptor extraction failure of some kind.
        :raises ValueError: Given data element content was not of a valid type
            with respect to this descriptor generator implementation.
        :raises IndexError: Underlying vector-producing generator either under
            or over produced vectors.

        :return: Iterator of result DescriptorElement instances. UUIDs of
            generated DescriptorElement instances will reflect the UUID of the
            DataElement it was generated from.
        :rtype: collections.Iterator[smqtk.representation.DescriptorElement]
        """
        log_debug = self._log.debug

        # Parallel lists of (uuid, DescriptorElement, already-computed) triples
        #   for formulating the return yielding.
        # Using deques so we can efficiently popleft off of them in the below
        #   for-loop. This way we do not retain elements and booleans for
        #   things we have yielded that would otherwise build up if this method
        #   iterated for a long time.
        #: :type: deque[(smqtk.representation.DescriptorElement, bool)]
        elem_and_status_q = deque()

        # Flag for end of data iteration. When not None will be the index of
        # the last data/descriptor element to be yielded. This will NOT be the
        # number of elements to be yielded, that would be ``end_of_iter[0]+1``.
        # NOTE: When moving to python3+ only support, we can use the
        #       ``nonlocal end_of_iter`` statement within the
        #       ``iter_tocompute_data`` function instead of imitating a state
        #       object here.
        #: :type: list[int|None]
        end_of_iter = [None]

        # TODO: Make generator threadsafe?
        # See: https://anandology.com/blog/using-iterators-and-generators/
        def iter_tocompute_data():
            """ Yield data elements that need descriptor computation.

            Populate parallel lists as we traverse ``data_iter``, yielding data
            elements that do not have an associated descriptor element with a
            vector set. Alternatively, if ``overwrite`` is True, all data
            elements are yielded and descriptor elements are marked for vector
            setting.

            :returns: Generator over data elements that need descriptor
                generation.
            :rtype: typing.Generator[smqtk.representation.DataElement]
            """
            # Running var for the index of final data element in input
            # iterator. This will be -1 or the value of the final index in the
            # parallel lists.
            last_i = -1
            for d_i, data in enumerate(data_iter):
                data_uuid_ = data.uuid()
                descr_elem_ = \
                    descr_factory.new_descriptor(self.name, data_uuid_)
                already_computed = not overwrite and descr_elem_.has_vector()
                elem_and_status_q.append((descr_elem_, already_computed))
                if not already_computed:
                    # Descriptor should be computed for this element
                    log_debug("Yielding DataElement with UUID {} for "
                              "generation".format(data_uuid_))
                    yield data
                else:
                    log_debug("Descriptor already computed for UUID {}"
                              .format(data_uuid_))
                last_i = d_i

            end_of_iter[0] = last_i

        descr_vec_iter = self.generate_arrays(iter_tocompute_data())
        for v_i, v in enumerate(descr_vec_iter):
            # These pops would fail with an IndexError if there is nothing left
            #   left from parallel allocation within ``iter_tocompute_data``.
            # This usually means that the ``self.generate_arrays`` is
            #   generating more vectors than there are descr element slots to
            #   fill.
            v_descr_elem, v_already_computed = elem_and_status_q.popleft()

            # Forwarding the iterator of the ``descr_vec_iter`` generator will,
            # probably, forward the ``iter_tocompute_data`` iterator, thus
            # populating the ``elem_and_status_q`` to some degree. The current
            # ``v`` should be be used to populate the next DescriptorElement
            # with an associated "already_computed" flag of False.
            while v_already_computed:
                yield v_descr_elem
                # We clearly have a descriptor vector from the result of
                # computation so there should logically be some future element
                # in which to store this result.
                v_descr_elem, v_already_computed = elem_and_status_q.popleft()

            # Assign the current computed descriptor vector to the current
            # element that should be set to.
            log_debug("Setting computed vector {} to element UUID {}"
                      .format(v_i, v_descr_elem.uuid()))
            v_descr_elem.set_vector(v)
            yield v_descr_elem

        # At this point, the ``iter_tocompute_data()`` iterator should have
        #   completed due to the ``self.generate_arrays`` method iterating
        #   through it completely, assigning a value to ``end_of_iter[0]``.
        # This also indicates that nothing more should be being added to the
        #   deques.
        assert end_of_iter[0] is not None, \
            "EoI value has not yet been assigned a value even though " \
            "``generate_arrays`` completed. Implementation may not have " \
            "fully iterated through the input DataElement iterable."

        # Finish yielding any already-computed descriptor elements that are
        # past the last computed element index.
        for e, already_comp in elem_and_status_q:
            # If an element's already computed state is False at this point,
            # then this implementation's ``_generate_arrays`` method must not
            # have yielded enough arrays to fill vacancies that needed to be
            # filled.
            if not already_comp:
                raise IndexError(
                    "Implementation generator under-produced vectors to fill "
                    "descriptor elements in need of a vector (UUID `{}` not "
                    "filled).".format(e.uuid())
                )
            yield e

    #
    # Single-element-based convenience methods
    #

    def generate_one_array(self, data_elem):
        """
        Convenience wrapper around ``generate_arrays`` for the single-input
        case.

        See the documentation for the
        :meth:`DescriptorGenerator.generate_arrays` method for more
        information.

        :param smqtk.representation.DataElement data_elem:
            DataElement instance to be described.

        :raises RuntimeError: Descriptor extraction failure of some kind.
        :raises ValueError: Given data element content was not of a valid type
            with respect to this descriptor generator implementation.

        :return: Descriptor vector the given data as a ``numpy.ndarray``
            instance.
        :rtype: numpy.ndarray
        """
        return list(
            self.generate_arrays([data_elem])
        )[0]

    def generate_one_element(self, data_elem,
                             descr_factory=DFLT_DESCRIPTOR_FACTORY,
                             overwrite=False):
        """
        Convenience wrapper around ``generate_elements`` for the single-input
        case.

        See documentation for the
        :meth:`DescriptorGenerator.generate_elements` method for more
        information

        :param smqtk.representation.DataElement data_elem:
            DataElement instance to be described.
        :param smqtk.representation.DescriptorElementFactory descr_factory:
            DescriptorElementFactory instance to drive the generation of
            element instances with some parametrization.
        :param bool overwrite:
            By default, if a factory-produced DescriptorElement reports as
            containing a vector, we do not compute a descriptor again for the
            associated data element. If this is ``True``, however, we will
            generate descriptors for all input data elements, overwriting the
            vectors previously stored in the factory-produces descriptor
            elements.
        :raises IndexError: Underlying vector-producing generator either under
            or over produced vectors.

        :raises RuntimeError: Descriptor extraction failure of some kind.
        :raises ValueError: Given data element content was not of a valid type
            with respect to this descriptor generator implementation.

        :return: Result DescriptorElement instance. UUID of the generated
            DescriptorElement instance will reflect the UUID of the
            DataElement it was generated from.
        :rtype: smqtk.representation.DescriptorElement
        """
        return list(
            self.generate_elements([data_elem], descr_factory=descr_factory,
                                   overwrite=overwrite)
        )[0]



"""
Data abstraction interface for general key-value storage.
"""
import abc
import collections

from smqtk.exceptions import ReadOnlyError
from smqtk.representation import SmqtkRepresentation
from smqtk.utils.plugin import Pluggable


NO_DEFAULT_VALUE = type("KeyValueStoreNoDefaultValueType", (object,), {})()


class KeyValueStore (SmqtkRepresentation, Pluggable):
    """
    Interface for general key/value storage.

    Implementations may impose restrictions on what types keys or values may be
    due to backend used.

    Data access and manipulation should be thread-safe.
    """

    # Mutable storage container is not hashable.
    __hash__ = None

    def __len__(self):
        return self.count()

    def __contains__(self, item):
        return self.has(item)

    def __getitem__(self, item):
        return self.get(item)

    @abc.abstractmethod
    def __repr__(self):
        """
        Return representative string for this class.

        *NOTE:* **This abstract super-method returns a template string to add
        to sub-class specific information to. The returned string should be
        formatted using the ``%`` operator and expects a single string
        argument.**

        :return: Representative string for this class.
        :rtype: str

        """
        return '<' + self.__class__.__name__ + " %s>"

    @abc.abstractmethod
    def count(self):
        """
        :return: The number of key-value relationships in this store.
        :rtype: int | long
        """

    @abc.abstractmethod
    def keys(self):
        """
        :return: Iterator over keys in this store.
        :rtype: collections.Iterator[collections.Hashable]
        """

    def values(self):
        """
        :return: Iterator over values in this store. Values are not guaranteed
            to be in any particular order.
        :rtype: collections.Iterator
        """
        for k in self.keys():
            yield self.get(k)

    @abc.abstractmethod
    def is_read_only(self):
        """
        :return: True if this instance is read-only and False if it is not.
        :rtype: bool
        """

    @abc.abstractmethod
    def has(self, key):
        """
        Check if this store has a value for the given key.

        :param key: Key to check for a value for.
        :type key: collections.Hashable

        :return: If this store has a value for the given key.
        :rtype: bool

        """

    @abc.abstractmethod
    def add(self, key, value):
        """
        Add a key-value pair to this store.

        *NOTE:* **Implementing sub-classes should call this super-method. This
        super method should not be considered a critical section for thread
        safety unless ``is_read_only`` is not thread-safe.**

        :param key: Key for the value. Must be hashable.
        :type key: collections.Hashable

        :param value: Python object to store.
        :type value: object

        :raises ReadOnlyError: If this instance is marked as read-only.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot add to read-only instance %s." % self)

    @abc.abstractmethod
    def add_many(self, d):
        """
        Add multiple key-value pairs at a time into this store as represented
        in the provided dictionary `d`.

        :param d: Dictionary of key-value pairs to add to this store.
        :type d: dict[collections.Hashable, object]

        :raises ReadOnlyError: If this instance is marked as read-only.

        :return: Self.
        :rtype: KeyValueStore

        """
        # Input keys must already be hashable because they're in a dictionary.
        if self.is_read_only():
            raise ReadOnlyError("Cannot add to read-only instance %s." % self)

    @abc.abstractmethod
    def remove(self, key):
        """
        Remove a single key-value entry.

        :param key: Key to remove.
        :type key: collections.Hashable

        :raises ReadOnlyError: If this instance is marked as read-only.
        :raises KeyError: The given key is not present in this store and no
            default value given.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot remove from read-only instance %s."
                                % self)

    @abc.abstractmethod
    def remove_many(self, keys):
        """
        Remove multiple keys and associated values.

        :param keys: Iterable of keys to remove.  If this is empty this method
            does nothing.
        :type keys: collections.Iterable[collections.Hashable]

        :raises ReadOnlyError: If this instance is marked as read-only.
        :raises KeyError: The given key is not present in this store and no
            default value given.  The store is not modified if any key is
            invalid.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot remove from read-only instance %s."
                                % self)

    @abc.abstractmethod
    def get(self, key, default=NO_DEFAULT_VALUE):
        """
        Get the value for the given key.

        *NOTE:* **Implementing sub-classes are responsible for raising a
        ``KeyError`` where appropriate.**

        :param key: Key to get the value of.
        :type key: collections.Hashable

        :param default: Optional default value if the given key is not present
            in this store. This may be any value except for the
            ``NO_DEFAULT_VALUE`` constant (custom anonymous class instance).
        :type default: object

        :raises KeyError: The given key is not present in this store and no
            default value given.

        :return: Deserialized python object stored for the given key.
        :rtype: object

        """

    def get_many(self, keys, default=NO_DEFAULT_VALUE):
        """
        Get the values for the given keys.

        *NOTE:* **Implementing sub-classes are responsible for raising a
        ``KeyError`` where appropriate.**

        :param keys: The keys for which associated values are requested.
        :type keys: collections.Iterable[collections.Hashable]

        :param default: Optional default value if a given key is not present
            in this store. This may be any value except for the
            ``NO_DEFAULT_VALUE`` constant (custom anonymous class instance).
        :type default: object

        :raises KeyError: A given key is not present in this store and no
            default value given.

        :return: Iterable of deserialized python objects stored for the given
            keys in the order that the corresponding keys were provided.
        :rtype: collections.Iterable

        """
        for key_ in keys:
            yield self.get(key_, default=default)

    @abc.abstractmethod
    def clear(self):
        """
        Clear this key-value store.

        *NOTE:* **Implementing sub-classes should call this super-method. This
        super method should not be considered a critical section for thread
        safety.**

        :raises ReadOnlyError: If this instance is marked as read-only.

        :return: Self.
        :rtype: KeyValueStore

        """
        if self.is_read_only():
            raise ReadOnlyError("Cannot clear a read-only %s instance."
                                % self.__class__.__name__)



# coding=utf-8
"""
Print out information about what plugins are currently usable and the
documentation headers for each implementation.
"""
from __future__ import print_function

import argparse
import json
import logging

import smqtk.algorithms
import smqtk.algorithms.nn_index.hash_index
import smqtk.algorithms.nn_index.lsh.functors
import smqtk.representation
import smqtk.utils.cli
from smqtk.utils.configuration import (
    make_default_config,
)


def cli():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-v", "--verbose",
                        default=False, action="store_true",
                        help="Output additional debug logging.")
    parser.add_argument("--defaults",
                        default=False, type=str,
                        help="Optionally generate default configuration blocks "
                             "for each plugin structure and output as JSON to "
                             "the specified path.")
    return parser


def format_cls_description(name, cls):
    return "%s\n%s" % (name, cls)


def main():
    args = cli().parse_args()

    llevel = logging.INFO
    if args.verbose:
        llevel = logging.DEBUG
    smqtk.utils.cli.initialize_logging(logging.getLogger("smqtk"), llevel)

    collect_defaults = args.defaults
    defaults = {}

    def collect_configs(name, impl_set):
        """
        :type name: str
        :type impl_set: set[type]
        """
        if collect_defaults:
            defaults[name] = make_default_config(impl_set)

    log = logging.getLogger("smqtk.checkPlugins")

    # Key is the interface type name
    #: :type: dict[str, set[type]]
    plugin_info = {}
    # List of plugin_info keys in order they were added
    plugin_type_list = []

    def collect_plugins(interface_t):
        """
        Record discoverable implementations for an interface class type.

        :param interface_t: Interface class type.

        """
        type_name = interface_t.__name__
        log.info("Checking %s plugins", type_name)
        plugin_type_list.append(type_name)
        #: :type: set[type]
        impl_set = interface_t.get_impls()
        plugin_info[plugin_type_list[-1]] = impl_set
        collect_configs(type_name, impl_set)

    #
    # smqtk.representation
    #
    collect_plugins(smqtk.representation.DataElement)
    collect_plugins(smqtk.representation.DataSet)
    collect_plugins(smqtk.representation.DescriptorElement)
    collect_plugins(smqtk.representation.DescriptorSet)
    collect_plugins(smqtk.representation.KeyValueStore)

    #
    # smqtk.algorithms
    #
    collect_plugins(smqtk.algorithms.Classifier)
    collect_plugins(smqtk.algorithms.DescriptorGenerator)
    collect_plugins(smqtk.algorithms.HashIndex)
    collect_plugins(smqtk.algorithms.LshFunctor)
    collect_plugins(smqtk.algorithms.NearestNeighborsIndex)
    collect_plugins(smqtk.algorithms.RelevancyIndex)

    #
    # Print-out
    #
    print()
    print()
    for k in plugin_type_list:
        print("[Type]", k)
        print('=' * (7 + len(k)))
        print()
        for t in plugin_info[k]:
            print(":: " + t.__name__)
            if t.__doc__:
                print(t.__doc__.rstrip().encode('utf-8'))
                print()
        print()
        print()

    if collect_defaults:
        with open(collect_defaults, 'w') as f:
            json.dump(defaults, f, indent=4, sort_keys=True,
                      separators=(',', ': '))
        log.info("Wrote default configuration dictionaries to: %s",
                 collect_defaults)


if __name__ == "__main__":
    main()



import abc
from collections import deque
import itertools

from six.moves import zip

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.representation import DescriptorElement

from ._defaults import DFLT_CLASSIFIER_FACTORY


class Classifier (SmqtkAlgorithm):
    """
    Interface for algorithms that classify input descriptors into discrete
    labels and/or label confidences.
    """

    @abc.abstractmethod
    def get_labels(self):
        """
        Get the sequence of class labels that this classifier can classify
        descriptors into. This includes the negative or background label if the
        classifier embodies such a concept.

        :return: Sequence of possible classifier labels.
        :rtype: collections.Sequence[collections.Hashable]

        :raises RuntimeError: No model loaded.

        """

    @abc.abstractmethod
    def _classify_arrays(self, array_iter):
        """
        Overridable method for classifying an iterable of descriptor elements
        whose vectors should be classified.

        At this level, all input arrays are guaranteed to be of consistent
        dimensionality.

        Each classification mapping should contain confidence values for each
        label the configured model contains.
        Implementations may act in a discrete manner whereby only one label is
        marked with a ``1`` value (others being ``0``), or in a continuous
        manner whereby each label is given a confidence-like value in the
        [0, 1] range.

        :param collections.Iterable[numpy.ndarray] array_iter:
            Iterable of arrays to be classified.

        :return: Iterable of dictionaries, parallel in association to the input
            descriptor vectors. Each dictionary should map labels to associated
            confidence values.
        :rtype: collections.Iterable[dict[collections.Hashable, float]]
        """

    @staticmethod
    def _assert_array_dim_consistency(array_iter):
        """
        Assert that arrays are consistent in dimensionality across iterated
        arrays.

        Currently we only support iterating single dimension vectors. Arrays
        of more than one dimension (i.e. 2D matries, etc.) will trigger a
        ValueError.

        :param collections.Iterable[numpy.ndarray] array_iter:
            Iterable numpy arrays.

        :raises ValueError: Not all input arrays were of consistent
            dimensionality.

        :return: Iterable of the same arrays in the same order, but validated
            to be of common dimensionality.
        """
        dim = None
        for a in array_iter:
            if a.ndim > 1:
                raise ValueError("Input vector had more than one dimension! "
                                 "(ndim = {})".format(a.ndim))
            elif dim is None:
                dim = a.size
            elif a.size != dim:
                raise ValueError("Input vector violated dimension consistency "
                                 "(basis == {}, violation == {})"
                                 .format(dim, a.size))
            yield a

    def classify_arrays(self, array_iter):
        """
        Classify an input iterable of numpy arrays into a parallel iterable of
        label-to-confidence mappings (dictionaries).

        Each classification mapping should contain confidence values for each
        label the configured model contains.
        Implementations may act in a discrete manner whereby only one label is
        marked with a ``1`` value (others being ``0``), or in a continuous
        manner whereby each label is given a confidence-like value in the
        [0, 1] range.

        :param collections.Iterable[numpy.ndarray] array_iter:
            Iterable of DescriptorElement instances to be classified.

        :raises ValueError: Input arrays were not all of consistent
            dimensionality.

        :return: Iterable of dictionaries, parallel in association to the input
            descriptor vectors. Each dictionary should map labels to associated
            confidence values.
        :rtype: collections.Iterable[dict[collections.Hashable, float]]
        """
        return self._classify_arrays(
            self._assert_array_dim_consistency(array_iter)
        )

    def classify_elements(self, descr_iter,
                          factory=DFLT_CLASSIFIER_FACTORY,
                          overwrite=False, d_elem_batch=100):
        """
        Classify an input iterable of descriptor elements into a parallel
        iterable of classification elements.

        Classification element UIDs are inherited from the descriptor element
        it was generated from.

        We invoke ``classify_arrays`` for actual generation of classification
        results. See documentation for this method for further details.
        # We invoke ``classify_arrays`` for factory-generated classification
        # elements that do not yet have classifications stored, or on all input
        # descriptor elements if the ``overwrite`` flag is True.

        **Selective Iteration**
        For situations when it is desired to access specific generator returns,
        like when only one descriptor element is provided in order to get a
        single element out, it is strongly recommended to expand the returned
        generator into a sequence type first.
        For example, expanding out the generator's returns into a list
        (``list(g.generate_elements([e]))[0]``) is recommended over just
        getting the "next" element of the returned generator
        (``next(g.generate_elements([e]))``).
        Expansion into a sequence allows the generator to fully execute, which
        includes any functionality after the final ``yield`` statement in any
        of the underlying iterators that may perform required clean-up.

        **Non-redundant Processing**
        Certain classification element implementations, as dictated by the
        input factory, may be connected to persistent storage in the
        background.
        Because of this, some classification elements may already "have"
        classification results on construction.
        This method, by default, only computes new classification results for
        descriptor elements whose associated classification element does not
        report as already containing results.
        If the ``overwrite`` flag is True then classifications are computed for
        all input descriptor elements and results are set to their respective
        classification elements regardless of existing result storage.

        :param collections.Iterable[DescriptorElement] descr_iter:
            Iterable of DescriptorElement instances to be classified.
        :param smqtk.representation.ClassificationElementFactory factory:
            Classification element factory. The default factory yields
            MemoryClassificationElement instances.
        :param bool overwrite:
            Recompute classification of the input descriptor and set the
            results to the ClassificationElement produced by the factory.
        :param int d_elem_batch:
            The number of descriptor elements to collect before requesting
            the whole batch's vectors at once via
            ``DescriptorElement.get_many_vectors`` method.

        :raises ValueError: Either: (A) one or more input descriptor elements
            did not have a stored vector, or (B) input descriptor element
            arrays were not all of consistent dimensionality.
        :raises IndexError: Implementation of ``_classify_arrays`` either under
            or over produced classifications relative to the number of input
            descriptor vectors.

        :return: Iterator of result ClassificationElement instances. UUIDs of
            generated ClassificationElement instances will reflect the UUID of
            the DescriptorElement it was computed from.
        :rtype: collections.Iterator[smqtk.representation.ClassificationElement]
        """
        log_debug = self._log.debug

        if d_elem_batch <= 0:
            self._log.warning("Descriptor element batching value <= 0, "
                              "defaulting to using value of 1.")
            d_elem_batch = 1

        # Queue populated by ``iter_tocompute_arrays`` with
        #   ClassificationElement instances paired with a flag indicating
        #   whether a classification was to be computed for that element.
        # Using deques so we can efficiently popleft off of them in the below
        #   for-loop. This way we do not retain elements and booleans for
        #   things we have yielded that would otherwise build up if this method
        #   iterated for a long time.
        #: :type: deque[(smqtk.representation.ClassificationElement, bool)]
        elem_and_status_q = deque()

        # Flag for end of data iteration. When not None will be the index of
        # the last descriptor/classification element to be yielded. This will
        # NOT be the number of elements to be yielded, that would be
        # ``end_of_iter[0]+1``.
        #: :type: list[int|None]
        end_of_iter = [None]

        # TODO: Make generator threadsafe?
        # See: https://anandology.com/blog/using-iterators-and-generators/
        def iter_tocompute_arrays():
            """ Yield descriptor vectors for classification elements that need
            computing yet.

            :rtype: typing.Generator[numpy.ndarray]
            """
            # Force into an iterator.
            descr_iterator = iter(descr_iter)
            # Running var for the index of final data element in input
            # iterator. This will be -1 or the value of the final index in the
            # parallel lists.
            last_i = -1
            # Make successive islices into iterator of descriptor elements to
            # produces batches. We end when there is nothing left being
            # returned by the iterator
            de_batch_list = \
                list(itertools.islice(descr_iterator, d_elem_batch))
            while de_batch_list:
                # Get vectors from batch using implementation-level batch
                # aggregation methods where applicable.
                de_batch_vecs = \
                    DescriptorElement.get_many_vectors(de_batch_list)

                for d_elem, d_vec in zip(de_batch_list, de_batch_vecs):
                    d_uid = d_elem.uuid()
                    if d_vec is None:
                        raise ValueError("Encountered DescriptorElement with "
                                         "no vector stored! (UID=`{}`)"
                                         .format(d_uid))
                    c_elem_ = factory.new_classification(self.name, d_uid)
                    already_computed = \
                        not overwrite and c_elem_.has_classifications()
                    elem_and_status_q.append((c_elem_, already_computed))
                    if not already_computed:
                        # Classifications should be computed for this
                        # descriptor
                        log_debug("Yielding descriptor array with UID `{}` "
                                  "for classification generation."
                                  .format(d_uid))
                        yield d_vec
                    else:
                        log_debug("Classification already generated for UID "
                                  "`{}`.".format(d_uid))

                last_i += len(de_batch_vecs)

                # Slice out the next batch of descriptor elements. This will be
                # empty if the iterator has been exhausted.
                de_batch_list = list(
                    itertools.islice(descr_iterator, d_elem_batch)
                )

            end_of_iter[0] = last_i

        classification_iter = self.classify_arrays(iter_tocompute_arrays())
        for c_i, c in enumerate(classification_iter):
            # These pops would fail with an IndexError if there is nothing left
            #   left from parallel allocation within ``iter_tocompute_arrays``.
            # This usually means that the ``self.generate_arrays`` is
            #   generating more vectors than there are descr element slots to
            #   fill.
            try:
                c_elem, c_already_computed = elem_and_status_q.popleft()
            except IndexError:
                # Translate index error into one with a more informative
                # message.
                raise IndexError(
                    "Implementation's ``_classify_arrays`` over-produced "
                    "classifications relative to input descriptor vectors."
                )

            # Forwarding the ``classification_iter`` generator should forward
            # the ``iter_tocompute_arrays`` iterator, thus populating the
            # ``elem_and_status_q`` by some amount. The current ``c`` should be
            # used to populate the next ClassificationElement with a False
            # ``already_computed`` flag. Yield classification elements that
            # already had classifications until we hit an element that was
            # flagged for computation.
            while c_already_computed:
                yield c_elem
                # We clearly have a classification from the result of
                # computation so there should logically be some future element
                # in which to store this result.
                c_elem, c_already_computed = elem_and_status_q.popleft()

            # We've arrived at an element that was flagged for computation, so
            # store the result.
            log_debug("Setting computed classification to element UID=`{}`"
                      .format(c_elem.uuid))
            c_elem.set_classification(c)
            yield c_elem

        # At this point, the ``iter_tocompute_arrays`` iterator should have
        #   completed du eto the ``self.classify_arrays`` method iterating
        #   through it completely, resulting in assignment to
        #   ``end_of_iter[0]``.
        # This also indicates that nothing more should be being added to
        #   ``elem_and_status_q``.
        assert end_of_iter[0] is not None, \
            "EoI value has not yet been assigned a value even though " \
            "``classify_arrays`` completed. Implementation of " \
            "``_classify_arrays`` may not have fully iterated through the " \
            "input numpy.ndarray iterable."

        # Finish yielding any "already-computed" classification elements that
        # are past the last computed element index.
        for c, already_comp in elem_and_status_q:
            # If an element's already-computed state is False at this point,
            # then the implementation's ``_classify_arrays`` method must not
            # have yielded enough arrays to fill the elements that were flagged
            # for classification.
            if not already_comp:
                raise IndexError(
                    "Implementation's ``_classify_arrays`` under-produced "
                    "classifications to fill elements that were flagged for "
                    "computation."
                )
            yield c

    def classify_one_element(self, descr_elem, factory=DFLT_CLASSIFIER_FACTORY,
                             overwrite=False):
        """
        Convenience method around ``classify_elements`` for the single-input
        case.

        See documentation for the :meth:`Classifier.classify_elements` method
        for more information.

        :param DescriptorElement descr_elem:
            Iterable of DescriptorElement instances to be classified.
        :param smqtk.representation.ClassificationElementFactory factory:
            Classification element factory. The default factory yields
            MemoryClassificationElement instances.
        :param bool overwrite:
            Recompute classification of the input descriptor and set the
            results to the ClassificationElement produced by the factory.

        :raises ValueError: The input descriptor element did not have a stored
            vector.
        :raises IndexError: Implementation of ``_classify_arrays`` either under
            or over produced classifications relative to the number of input
            descriptor vectors.

        :return: ClassificationElement instances. UUIDs of the generated
            ClassificationElement instance will reflect the UUID of the
            DescriptorElement it was computed from.
        :rtype: smqtk.representation.ClassificationElement
        """
        return list(self.classify_elements(
            [descr_elem], factory=factory, overwrite=overwrite, d_elem_batch=1
        ))[0]



"""
SMQTK Web Applications
"""

import inspect
import os

import flask

from smqtk.utils import SmqtkObject
from smqtk.utils.configuration import Configurable
from smqtk.utils.dict import merge_dict
from smqtk.utils.plugin import Pluggable


# noinspection PyAbstractClass
class SmqtkWebApp (SmqtkObject, flask.Flask, Configurable, Pluggable):
    """
    Base class for SMQTK web applications
    """

    @classmethod
    def impl_directory(cls):
        """
        :return: Directory in which this implementation is contained.
        :rtype: str
        """
        return os.path.dirname(os.path.abspath(inspect.getfile(cls)))

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        This should be overridden in each implemented application class to add
        appropriate configuration.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        return {
            "flask_app": {
                "SECRET_KEY": "MySuperUltraSecret",
                "BASIC_AUTH_USERNAME": "demo",
                "BASIC_AUTH_PASSWORD": "demo"
            },
            "server": {
                'host': "127.0.0.1",
                'port': 5000
            }
        }

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Override to just pass the configuration dictionary to constructor
        """
        # Repeated from super method due to overriding how constructor is called
        if merge_default:
            merged = cls.get_default_config()
            merge_dict(merged, config_dict)
            config_dict = merged
        return cls(config_dict)

    def __init__(self, json_config):
        """
        Initialize application based of supplied JSON configuration

        :param json_config: JSON configuration dictionary
        :type json_config: dict

        """
        super(SmqtkWebApp, self).__init__(
            self.__class__.__name__,
            static_folder=os.path.join(self.impl_directory(), 'static'),
            template_folder=os.path.join(self.impl_directory(), 'templates')
        )

        #
        # Configuration setup
        #
        self.json_config = json_config

        # Factor 'flask_app' configuration properties into self.config
        for k in self.json_config['flask_app']:
            self.config[k] = self.json_config['flask_app'][k]

        #
        # Security
        #
        self.secret_key = self.config['SECRET_KEY']

    def get_config(self):
        return self.json_config

    def run(self, host=None, port=None, debug=False, **options):
        """
        Override of the run method, drawing running host and port from
        configuration by default. 'host' and 'port' values specified as argument
        or keyword will override the app configuration.
        """
        super(SmqtkWebApp, self)\
            .run(host=(host or self.json_config['server']['host']),
                 port=(port or self.json_config['server']['port']),
                 debug=debug,
                 **options)



from ._interface import SmqtkAlgorithm

# Import module abstracts and plugin getter functions
from .classifier import Classifier, SupervisedClassifier
from .descriptor_generator import DescriptorGenerator
from .image_io import ImageReader
from .object_detection import ObjectDetector, ImageMatrixObjectDetector
from .nn_index import NearestNeighborsIndex
from .nn_index.hash_index import HashIndex
from .nn_index.lsh.functors import LshFunctor
from .relevancy_index import RelevancyIndex



from ._interface import SmqtkRepresentation

from .bbox import AxisAlignedBoundingBox

from .classification_element import ClassificationElement
from .data_element import DataElement
from .data_set import DataSet
from .descriptor_element import DescriptorElement
from .descriptor_set import DescriptorSet
from .detection_element import DetectionElement
from .key_value import KeyValueStore

from .classification_element_factory import ClassificationElementFactory
from .descriptor_element_factory import DescriptorElementFactory
from .detection_element_factory import DetectionElementFactory



# noinspection PyUnresolvedReferences
# - Convenience pass-through from "private" sub-module.
from ._defaults import DFLT_CLASSIFIER_FACTORY, DFLT_DETECTION_FACTORY
from ._interface import ObjectDetector, ImageMatrixObjectDetector



import itertools
import logging

import numpy
import PIL.Image
import PIL.ImageFile
from six import BytesIO
from six.moves import zip

from smqtk.algorithms.descriptor_generator import DescriptorGenerator
from smqtk.representation import DataElement
from smqtk.utils.configuration import from_config_dict, to_config_dict, \
    make_default_config
from smqtk.utils.dict import merge_dict
from smqtk.utils.parallel import parallel_map

try:
    import caffe
except ImportError as ex:
    logging.getLogger(__name__).warning("Failed to import caffe module: %s",
                                        str(ex))
    caffe = None


__author__ = 'paul.tunison@kitware.com, jacob.becker@kitware.com'

__all__ = [
    "CaffeDescriptorGenerator",
]


class CaffeDescriptorGenerator (DescriptorGenerator):
    """
    Compute images against a Caffe model, extracting a layer as the content
    descriptor.
    """

    @classmethod
    def is_usable(cls):
        valid = caffe is not None
        if not valid:
            cls.get_logger().debug("Caffe python module cannot be imported")
        return valid

    @classmethod
    def get_default_config(cls):
        default = super(CaffeDescriptorGenerator, cls).get_default_config()

        data_elem_impl_set = DataElement.get_impls()
        # Need to make copies of dict so changes to one does not effect others.
        default['network_prototxt'] = \
            make_default_config(data_elem_impl_set)
        default['network_model'] = make_default_config(data_elem_impl_set)
        default['image_mean'] = make_default_config(data_elem_impl_set)

        return default

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(),
                                     config_dict)

        data_elem_impl_set = DataElement.get_impls()

        # Translate prototext and model sub-configs into DataElement instances.
        config_dict['network_prototxt'] = \
            from_config_dict(config_dict['network_prototxt'],
                             data_elem_impl_set)
        config_dict['network_model'] = \
            from_config_dict(config_dict['network_model'],
                             data_elem_impl_set)

        # Translate optionally provided image mean sub-config into a
        # DataElement instance. May have been provided as ``None`` or a
        # configuration dictionary with type ``None`.
        # None, dict[type=None], dict[type=str]
        if config_dict['image_mean'] is None \
                or config_dict['image_mean'].get('type', None) is None:
            config_dict['image_mean'] = None
        else:
            config_dict['image_mean'] = \
                from_config_dict(config_dict['image_mean'], data_elem_impl_set)

        return super(CaffeDescriptorGenerator, cls)\
            .from_config(config_dict, merge_default=False)

    def __init__(self, network_prototxt, network_model,
                 image_mean=None, return_layer='fc7',
                 batch_size=1, use_gpu=False, gpu_device_id=0,
                 network_is_bgr=True, data_layer='data',
                 load_truncated_images=False, pixel_rescale=None,
                 input_scale=None, threads=None):
        """
        Create a Caffe CNN descriptor generator

        :param smqtk.representation.DataElement network_prototxt: Data element
            containing the text file defining the network layout.

        :param smqtk.representation.DataElement network_model: Data element
            containing the trained ``.caffemodel`` file to use.

        :param smqtk.representation.DataElement image_mean: Optional data
            element containing the image mean ``.binaryproto`` or ``.npy``
            file.

        :param return_layer: The label of the layer we take data from to compose
            output descriptor vector.
        :type return_layer: str

        :param batch_size: The maximum number of images to process in one feed
            forward of the network. This is especially important for GPUs since
            they can only process a batch that will fit in the GPU memory space.
        :type batch_size: int

        :param use_gpu: If Caffe should try to use the GPU
        :type use_gpu: bool

        :param gpu_device_id: Integer ID of the GPU device to use. Only used if
            ``use_gpu`` is True.
        :type gpu_device_id: int

        :param network_is_bgr: If the network is expecting BGR format pixels.
            For example, the BVLC default caffenet does (thus the default is
            True).
        :type network_is_bgr: bool

        :param data_layer: String label of the network's data layer.
            We assume its 'data' by default.
        :type data_layer: str

        :param load_truncated_images: If we should be lenient and force loading
            of truncated image bytes. This is False by default.
        :type load_truncated_images: bool

        :param pixel_rescale: Re-scale image pixel values before being
            transformed by caffe (before mean subtraction, etc)
            into the given tuple ``(min, max)`` range. By default, images are
            loaded in the ``[0, 255]`` range. Refer to the image mean being used
            for desired input pixel scale.
        :type pixel_rescale: None | (float, float)

        :param input_scale: Optional floating-point scalar value to scale values
            of caffe network input data AFTER mean subtraction. This value is
            directly multiplied against the pixel values.
        :type input_scale: None | float

        :param int|None threads:
            Optional specific number of threads to use for data loading and
            pre-processing. If this is None or 0, we introspect the current
            system thread capacity and use that.

        ::raises AssertionError: Optionally provided image mean protobuf
            consisted of more than one image, or its shape was neither 1 or 3
            channels.
        """
        super(CaffeDescriptorGenerator, self).__init__()

        self.network_prototxt = network_prototxt
        self.network_model = network_model
        self.image_mean = image_mean

        self.return_layer = str(return_layer)
        self.batch_size = int(batch_size)

        self.use_gpu = bool(use_gpu)
        self.gpu_device_id = int(gpu_device_id)

        self.network_is_bgr = bool(network_is_bgr)
        self.data_layer = str(data_layer)

        self.load_truncated_images = bool(load_truncated_images)
        self.pixel_rescale = pixel_rescale
        self.input_scale = input_scale

        self.threads = threads

        assert self.batch_size > 0, \
            "Batch size must be greater than 0 (got %d)" \
            % self.batch_size
        assert self.gpu_device_id >= 0, \
            "GPU Device ID must be greater than 0 (got %d)" \
            % self. gpu_device_id

        # Network setup variables
        self.network = None
        self.net_data_shape = ()
        self.transformer = None

        self._setup_network()

    def __getstate__(self):
        return self.get_config()

    def __setstate__(self, state):
        # This ``__dict__.update`` works because configuration parameters
        # exactly match up with instance attributes currently.
        self.__dict__.update(state)
        # Translate nested Configurable instance configurations into actual
        # object instances.
        # noinspection PyTypeChecker
        self.network_prototxt = from_config_dict(
            self.network_prototxt, DataElement.get_impls()
        )
        # noinspection PyTypeChecker
        self.network_model = from_config_dict(
            self.network_model, DataElement.get_impls()
        )
        if self.image_mean is not None:
            # noinspection PyTypeChecker
            self.image_mean = from_config_dict(
                self.image_mean, DataElement.get_impls()
            )
        self._setup_network()

    def _set_caffe_mode(self):
        """
        Set the appropriate Caffe mode on the current thread/process.
        """
        if self.use_gpu:
            self._log.debug("Using GPU")
            caffe.set_device(self.gpu_device_id)
            caffe.set_mode_gpu()
        else:
            self._log.debug("using CPU")
            caffe.set_mode_cpu()

    def _setup_network(self):
        """
        Initialize Caffe and the network

        ::raises AssertionError: Optionally provided image mean protobuf
            consisted of more than one image, or its shape was neither 1 or 3
            channels.
        """
        self._set_caffe_mode()

        # Questions:
        #   - ``caffe.TEST`` indicates phase of either TRAIN or TEST
        self._log.debug("Initializing network")
        self._log.debug("Loading Caffe network from network/model configs")
        self.network = caffe.Net(
            self.network_prototxt.write_temp().encode(),
            caffe.TEST,
            weights=self.network_model.write_temp().encode()
        )
        self.network_prototxt.clean_temp()
        self.network_model.clean_temp()
        # Assuming the network has a 'data' layer and notion of data shape
        self.net_data_shape = self.network.blobs[self.data_layer].data.shape
        self._log.debug("Network data shape: %s", self.net_data_shape)

        # Crating input data transformer
        self._log.debug("Initializing data transformer")
        self.transformer = caffe.io.Transformer(
            {self.data_layer: self.network.blobs[self.data_layer].data.shape}
        )
        self._log.debug("Initializing data transformer -> %s",
                        self.transformer.inputs)

        if self.image_mean is not None:
            self._log.debug("Loading image mean (reducing to single pixel "
                            "mean)")
            image_mean_bytes = self.image_mean.get_bytes()
            try:
                # noinspection PyTypeChecker
                a = numpy.load(BytesIO(image_mean_bytes), allow_pickle=True)
                self._log.info("Loaded image mean from numpy bytes")
            except IOError:
                self._log.debug("Image mean file not a numpy array, assuming "
                                "URI to protobuf binary.")
                # noinspection PyUnresolvedReferences
                blob = caffe.proto.caffe_pb2.BlobProto()
                blob.ParseFromString(image_mean_bytes)
                a = numpy.array(caffe.io.blobproto_to_array(blob))
                assert a.shape[0] == 1, \
                    "Input image mean blob protobuf consisted of more than " \
                    "one image. Not sure how to handle this yet."
                a = a.reshape(a.shape[1:])
                self._log.info("Loaded image mean from protobuf bytes")
            assert a.shape[0] in [1, 3], \
                "Currently asserting that we either get 1 or 3 channel " \
                "images. Got a %d channel image." % a[0]
            # TODO: Instead of always using pixel mean, try to use image-mean
            #       if given. Might have to rescale if image/data layer shape
            #       is different.
            a_mean = a.mean(1).mean(1)
            self._log.debug("Initializing data transformer -- mean")
            self.transformer.set_mean(self.data_layer, a_mean)

        self._log.debug("Initializing data transformer -- transpose")
        self.transformer.set_transpose(self.data_layer, (2, 0, 1))
        if self.network_is_bgr:
            self._log.debug("Initializing data transformer -- channel swap")
            self.transformer.set_channel_swap(self.data_layer, (2, 1, 0))
        if self.input_scale:
            self._log.debug("Initializing data transformer -- input scale")
            self.transformer.set_input_scale(self.data_layer, self.input_scale)

    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        In the common case, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        if self.image_mean is not None:
            image_mean_config = to_config_dict(self.image_mean)
        else:
            image_mean_config = None
        return {
            "network_prototxt": to_config_dict(self.network_prototxt),
            "network_model": to_config_dict(self.network_model),
            "image_mean": image_mean_config,
            "return_layer": self.return_layer,
            "batch_size": self.batch_size,
            "use_gpu": self.use_gpu,
            "gpu_device_id": self.gpu_device_id,
            "network_is_bgr": self.network_is_bgr,
            "data_layer": self.data_layer,
            "load_truncated_images": self.load_truncated_images,
            "pixel_rescale": self.pixel_rescale,
            "input_scale": self.input_scale,
            "threads": self.threads,
        }

    def valid_content_types(self):
        """
        :return: A set valid MIME type content types that this descriptor can
            handle.
        :rtype: set[str]
        """
        return {
            'image/tiff',
            'image/png',
            'image/jpeg',
            'image/bmp',
        }

    def _generate_arrays(self, data_iter):
        """
        Inner template method that defines the generation of descriptor vectors
        for a given iterable of data elements.

        Pre-conditions:
          - Data elements input to this method have been validated to be of at
            least one of this class's reported ``valid_content_types``.

        :param collections.Iterable[DataElement] data_iter:
            Iterable of data element instances to be described.

        :raises RuntimeError: Descriptor extraction failure of some kind.

        :return: Iterable of numpy arrays in parallel association with the
            input data elements.
        :rtype: collections.Iterable[numpy.ndarray]
        """
        self._set_caffe_mode()
        log_debug = self._log.debug

        # Start parallel operation to pre-process imagery before aggregating
        # for network execution.
        # TODO: update ``buffer_factor`` param to account for batch size?
        img_array_iter = \
            parallel_map(_process_load_img_array,
                         zip(
                             data_iter, itertools.repeat(self.transformer),
                             itertools.repeat(self.data_layer),
                             itertools.repeat(self.load_truncated_images),
                             itertools.repeat(self.pixel_rescale),
                         ),
                         ordered=True, cores=self.threads)

        # Aggregate and process batches of input data elements
        #: :type: list[numpy.ndarray]
        batch_img_arrays = \
            list(itertools.islice(img_array_iter, self.batch_size))
        batch_i = 0
        while len(batch_img_arrays) > 0:
            cur_batch_size = len(batch_img_arrays)
            log_debug("Batch {} - size {}".format(batch_i, cur_batch_size))

            log_debug("Updating network data layer shape ({} images)"
                      .format(cur_batch_size))
            self.network.blobs[self.data_layer].reshape(
                cur_batch_size, *self.net_data_shape[1:4]
            )
            log_debug("Loading image matrices into network layer '{:s}'"
                      .format(self.data_layer))
            self.network.blobs[self.data_layer].data[...] = batch_img_arrays
            log_debug("Moving network forward")
            self.network.forward()
            descriptor_list = self.network.blobs[self.return_layer].data
            log_debug("extracting return layer '{:s}' into vectors"
                      .format(self.return_layer))
            for v in descriptor_list:
                if v.ndim > 1:
                    # In case caffe generates multidimensional array
                    # (rows, 1, 1)
                    log_debug("- Raveling output array of shape {}"
                              .format(v.shape))
                    yield numpy.ravel(v)
                else:
                    yield v

            # Slice out the next batch
            #: :type: list[(collections.Hashable, numpy.ndarray)]
            batch_img_arrays = \
                list(itertools.islice(img_array_iter, self.batch_size))
            batch_i += 1


def _process_load_img_array(input_tuple):
    """
    Helper function for multiprocessing image data loading

    Expected input argument tuple contents (in tuple order):
        * data_element: DataElement providing bytes
        * transformer: Caffe Transformer instance for pre-processing.
        * data_layer: String label of the network's data layer
        * load_truncated_images: Boolean of whether loading truncated images is
          allowed (See PIL.ImageFile.LOAD_TRUNCATED_IMAGES attribute).
        * pixel_rescale: Pair of floating point values to recale image values
          into, i.e. [0, 255] (the default).

    :param input_tuple:
        Tuple of input arguments as we expect to be called by a multiprocessing
        map function. See above for content details.

    :return: Input DataElement UUID and Pre-processed numpy array.
    :rtype: (collections.Hashable, numpy.ndarray)

    """
    # data_element: DataElement providing bytes
    # transformer: Caffe Transformer instance for pre-processing.
    # data_layer: String label of the data layer
    (data_element, transformer, data_layer, load_truncated_images,
     pixel_rescale) = input_tuple
    PIL.ImageFile.LOAD_TRUNCATED_IMAGES = load_truncated_images
    try:
        img = PIL.Image.open(BytesIO(data_element.get_bytes()))
    except Exception as ex:
        logging.getLogger(__name__).error(
            "Failed opening image from data element {}. Exception ({}): {}"
            .format(data_element, type(ex), str(ex))
        )
        raise
    if img.mode != "RGB":
        img = img.convert("RGB")
    logging.getLogger(__name__).debug("Image: {}".format(img))
    # Caffe natively uses float types (32-bit)
    try:
        # This can fail if the image is truncated and we're not allowing the
        # loading of those images
        img_a = numpy.asarray(img, numpy.float32)
    except Exception as ex:
        logging.getLogger(__name__).error(
            "Failed array-ifying data element {}. Image may be truncated. "
            "Exception ({}): {}"
            .format(data_element, type(ex), str(ex))
        )
        raise
    assert img_a.ndim == 3, \
        "Loaded invalid RGB image with shape {:s}".format(img_a.shape)
    if pixel_rescale:
        pmin, pmax = min(pixel_rescale), max(pixel_rescale)
        r = pmax - pmin
        img_a = (img_a / (255. / r)) + pmin
    img_at = transformer.preprocess(data_layer, img_a)
    return img_at



"""
Script for building and saving the model for the ``SkLearnBallTreeHashIndex``
implementation of ``HashIndex``.
"""
import logging

from smqtk.algorithms.nn_index.hash_index.sklearn_balltree import \
    SkLearnBallTreeHashIndex
import smqtk.representation
import smqtk.utils.cli
import smqtk.utils.bits
from smqtk.utils.configuration import (
    from_config_dict,
    make_default_config,
)


def default_config():
    return {
        "hash2uuid_kv_store": make_default_config(
            smqtk.representation.KeyValueStore.get_impls()
        ),
        "sklearn_balltree": SkLearnBallTreeHashIndex.get_default_config(),
        "itq_bit_length": 256,
    }


def cli_parser():
    parser = smqtk.utils.cli.basic_cli_parser(__doc__)
    return parser


def main():
    args = cli_parser().parse_args()
    config = smqtk.utils.cli.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    # Loading from configurations
    bit_len = int(config['itq_bit_length'])
    log.info("Loading hash2uuid KeyValue store")
    #: :type: smqtk.representation.KeyValueStore
    hash2uuid_kv_store = from_config_dict(
        config['hash2uuid_kv_store'],
        smqtk.representation.KeyValueStore.get_impls()
    )
    log.info("Initializing ball tree")
    btree = SkLearnBallTreeHashIndex.from_config(config['sklearn_balltree'])

    log.info("Computing hash-code vectors")
    hash_vectors = []
    reporter = smqtk.utils.cli.ProgressReporter(log.debug, 1.0)
    reporter.start()
    for h in hash2uuid_kv_store.keys():
        # Casting to int is a valid thing to do.
        # noinspection PyTypeChecker
        hash_vectors.append(
            smqtk.utils.bits.int_to_bit_vector_large(int(h), bit_len))
        reporter.increment_report()
    reporter.report()

    log.info("Building ball tree index")
    btree.build_index(hash_vectors)


if __name__ == '__main__':
    main()



import abc
import collections

from smqtk.representation import SmqtkRepresentation
from smqtk.utils.plugin import Pluggable


class DataSet (collections.Set, SmqtkRepresentation, Pluggable):
    """
    Abstract interface for data sets, that contain an arbitrary number of
    ``DataElement`` instances of arbitrary implementation type, keyed on
    ``DataElement`` UUID values.

    This should only be used with DataElements whose byte content is expected
    not to change. If they do, then UUID keys may no longer represent the
    elements associated with them.

    """

    def __len__(self):
        """
        :return: Number of elements in this DataSet.
        :rtype: int
        """
        return self.count()

    def __getitem__(self, uuid):
        return self.get_data(uuid)

    def __contains__(self, d):
        """
        Different than has_uuid() because this takes another DataElement
        instance, not a UUID.

        :param d: DataElement to test for containment
        :type d: smqtk.representation.DataElement

        :return: True of this DataSet contains the given data element. Since,
        :rtype: bool

        """
        return self.has_uuid(d.uuid())

    @abc.abstractmethod
    def __iter__(self):
        """
        :return: Generator over the DataElements contained in this set in no
            particular order.
        """

    @abc.abstractmethod
    def count(self):
        """
        :return: The number of data elements in this set.
        :rtype: int
        """

    @abc.abstractmethod
    def uuids(self):
        """
        :return: A new set of uuids represented in this data set.
        :rtype: set
        """

    @abc.abstractmethod
    def has_uuid(self, uuid):
        """
        Test if the given uuid refers to an element in this data set.

        :param uuid: Unique ID to test for inclusion. This should match the
            type that the set implementation expects or cares about.
        :type uuid: collections.Hashable

        :return: True if the given uuid matches an element in this set, or
            False if it does not.
        :rtype: bool

        """

    @abc.abstractmethod
    def add_data(self, *elems):
        """
        Add the given data element(s) instance to this data set.

        *NOTE: Implementing methods should check that input elements are in
        fact DataElement instances.*

        :param elems: Data element(s) to add
        :type elems: smqtk.representation.DataElement

        """

    @abc.abstractmethod
    def get_data(self, uuid):
        """
        Get the data element the given uuid references, or raise an
        exception if the uuid does not reference any element in this set.

        :raises KeyError: If the given uuid does not refer to an element in
            this data set.

        :param uuid: The uuid of the element to retrieve.
        :type uuid: collections.Hashable

        :return: The data element instance for the given uuid.
        :rtype: smqtk.representation.DataElement

        """



"""
Mongo Session wrapper

Allows session updating within AJAX routines.

Taken from: http://flask.pocoo.org/snippets/110/

"""

from datetime import datetime, timedelta
import logging
from uuid import uuid4

from flask.sessions import SessionInterface, SessionMixin
from werkzeug.datastructures import CallbackDict
from pymongo import MongoClient


LOG = logging.getLogger(__name__)


class MongoSession(CallbackDict, SessionMixin):

    def __init__(self, initial=None, sid=None):
        super(MongoSession, self).__init__(initial, MongoSession.on_update)
        self.sid = sid
        self.modified = False

    def on_update(self):
        self.modified = True


class MongoSessionInterface(SessionInterface):

    def __init__(self, host='localhost', port=27017,
                 db='', collection='sessions', delete_on_empty=False):
        client = MongoClient(host, port)
        self.store = client[db][collection]
        self._delete_on_empty = delete_on_empty

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        if sid:
            stored_session = self.store.find_one({'_id': sid})
            if stored_session:
                if stored_session.get('expiration') > datetime.utcnow():
                    LOG.debug("Returning existing MongoSession instance for "
                              "SID={}".format(sid))
                    return MongoSession(initial=stored_session['data'],
                                        sid=stored_session['_id'])
        sid = str(uuid4())
        LOG.debug("Returning NEW MongoSession instance for SID={}"
                  .format(sid))
        return MongoSession(sid=sid)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        if self._delete_on_empty and not session:
            LOG.debug("Session cookie content was empty, deleting.")
            response.delete_cookie(app.session_cookie_name, domain=domain)
            # TODO: Delete from mongo as well?
            return
        # Update expiration due to access.
        expiration = self.get_expiration_time(app, session)
        if not expiration:
            expiration = datetime.utcnow() + timedelta(hours=1)
        self.store.update({'_id': session.sid},
                          {'data': session,
                           'expiration': expiration},
                          upsert=True)
        LOG.debug("Setting session cookie for SID={}".format(session.sid))
        response.set_cookie(app.session_cookie_name, session.sid,
                            expires=expiration,
                            httponly=True, domain=domain)



import functools
import math
import operator as op

from six.moves import range


def ncr(n, r):
    """
    N-choose-r method, returning the number of combinations possible in integer
    form.

    From dheerosaur:
        http://stackoverflow.com/questions/4941753/is-there-a-math-ncr-function-in-python

    :param n: Selection pool size.
    :type n: int

    :param r: permutation selection size.
    :type r: int

    :return: Number of n-choose-r permutations for the given n and r.
    :rtype: int

    """
    r = min(r, n - r)
    if r == 0:
        return 1
    numer = functools.reduce(op.mul, range(n, n - r, -1), 1)
    # denom = functools.reduce(op.mul, range(1, r+1), 1)
    denom = math.factorial(r)
    return numer // denom





import abc

from smqtk.algorithms import SmqtkAlgorithm


class NoIndexError (Exception):
    """
    When a RelevancyIndex does not yet have a index built yet when one is
    needed.
    """


class RelevancyIndex (SmqtkAlgorithm):
    """
    Abstract class for IQR index implementations.

    Similar to a traditional nearest-neighbors algorithm, An IQR index provides
    a specialized nearest-neighbors interface that can take multiple examples of
    positively and negatively relevant exemplars in order to produce a [0, 1]
    ranking of the indexed elements by determined relevancy.

    """

    def __len__(self):
        return self.count()

    @abc.abstractmethod
    def count(self):
        """
        :return: Number of elements in this index.
        :rtype: int
        """

    @abc.abstractmethod
    def build_index(self, descriptors):
        """
        Build the index based on the given iterable of descriptor elements.

        Subsequent calls to this method should rebuild the index, not add to it.

        :raises ValueError: No data available in the given iterable.

        :param descriptors: Iterable of descriptor elements to build index over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """

    @abc.abstractmethod
    def rank(self, pos, neg):
        """
        Rank the currently indexed elements given ``pos`` positive and ``neg``
        negative exemplar descriptor elements.

        :param pos: Iterable of positive exemplar DescriptorElement instances.
            This may be optional for some implementations.
        :type pos: collections.Iterable[smqtk.representation.DescriptorElement]

        :param neg: Iterable of negative exemplar DescriptorElement instances.
            This may be optional for some implementations.
        :type neg: collections.Iterable[smqtk.representation.DescriptorElement]

        :raises NoIndexError:
            If index ranking is requested without an index to rank.

        :return: Map of indexed descriptor elements to a rank value between
            [0, 1] (inclusive) range, where a 1.0 means most relevant and 0.0
            meaning least relevant.
        :rtype: dict[smqtk.representation.DescriptorElement, float]

        """



"""
Helper interface and function for implementing class discovery.

Plugins may be accessed by one of the following ways:
  1. Be defined within SMQTK next to their interface.
  2. The environment variable ``SMQTK_PLUGIN_PATH`` may be set to a number of
     `:`-separated (`;` on Windows) python module paths to where plugin classes
     are defined.
  3. Other installed python packages may define one or more extensions for
     the namespace "smqtk_plugins". This should be a single or list of
     extensions that specify modules within the installed package where
     plugins for export are implemented. Note that we desire modules,
     not objects, for our extensions.

     For example::

         ...
         entry_points={
             "smqtk_plugins": "my_package = my_package.plugins"
         ]
         ...

     Or::

         ...
         entry_points = {
             "smqtk_plugins": [
                 "my_package_mode_1 = my_package.mode_1.plugins",
                 "my_package_mode_2 = my_package.mode_2.plugins",
             ]
         }
         ...

"""

import abc
import collections
import importlib
import inspect
import itertools
import logging
import os
import pkgutil
import re
import types
import warnings

import six
from six.moves import reload_module
from stevedore.extension import ExtensionManager


# Template for checking validity of sub-module files
VALID_MODULE_FILE_RE = re.compile(r"^[a-zA-Z]\w*(?:\.py)?$")

# Template for checking validity of module attributes
VALUE_ATTRIBUTE_RE = re.compile(r"^[a-zA-Z]\w*$")

# Environment variable *PATH separator for the current platform.
OS_ENV_PATH_SEP = os.pathsep

EXTENSION_NAMESPACE = "smqtk_plugins"


def _get_local_plugin_modules(log, interface_type, warn=True):
    """
    Get the python modules within the SMQTK tree that are located parallel
    to the module defining the given interface type.

    :param logging.Logger log:
        Logger instance to use for logging.
    :param type interface_type:
        Interface type we want to find modules around.
    :param bool warn:
        If we should warn about module import failures.

    :return: Iterator of python modules parallel to the given interface.
    :rtype: collections.Iterator[types.ModuleType]

    """
    # Get the parent module and the filesystem path to that module.
    # - This is for finding implementations that are defined in the same
    #   parent module as the interface that is inheriting from
    #   ``Pluggable``.
    # - This should do the right thing regardless of whether ``cls`` is
    #   defined in an ``__init__.py`` file or "normal" python module file.
    #: :type: types.ModuleType
    t_module = inspect.getmodule(interface_type)
    # __package__ should correctly get us the parent module path regardless of
    # whether ``interface_type`` is defined in an ``__init__.py`` file or in a
    # parallel python module file.
    #: :type: str
    t_module_package = t_module.__package__

    # Containing directory of the module ``interface_class`` is defined in.
    t_module_fp = inspect.getsourcefile(t_module) or inspect.getfile(t_module)
    t_module_dir = os.path.abspath(os.path.dirname(t_module_fp))
    log.debug("Looking for python modules parallel to {} in directory '{}'."
              .format(interface_type.__name__, t_module_dir))

    # Discover sibling modules to interface type's module.
    for importer, module_name, ispackage \
            in pkgutil.iter_modules([t_module_dir]):
        sib_module_path = '.'.join([t_module_package, module_name])
        log.debug("Found sibling module '{}'.".format(sib_module_path))
        try:
            yield importlib.import_module(sib_module_path)
        except Exception as ex:
            if warn:
                warnings.warn("Failed to import module '{}' due to exception: "
                              "({}) {}"
                              .format(sib_module_path, ex.__class__.__name__,
                                      str(ex)))


def _get_envvar_plugin_module(log, env_var, warn=True):
    """
    Get the python modules specified by the given environment variable.

    :param logging.Logger log:
        Logger instance to use for logging.
    :param str env_var:
        Environment variable key to use to look for python module paths to
        load.
    :param bool warn:
        If we should warn about module import failures.

    :return: Iterator of python modules parallel to the given interface.
    :rtype: collections.Iterator[types.ModuleType]

    """
    if env_var in os.environ:
        for p in os.environ[env_var].split(OS_ENV_PATH_SEP):
            # skip empty strings
            if p:
                log.debug("In env variable '{}' found module path '{}'."
                          .format(env_var, p))
                try:
                    yield importlib.import_module(p)
                except Exception as ex:
                    if warn:
                        warnings.warn(
                            "Failed to import module '{}' due to exception: "
                            "({}) {}"
                            .format(p, ex.__class__.__name__, str(ex))
                        )
    else:
        log.debug("No variable '{}' in environment.".format(env_var))


def _get_extension_plugin_modules(log, warn=True):
    """
    Get the modules registered by installed python modules that
    provide extensions in the ``smqtk_plugins`` namespace.

    :return: Iterator of python modules registered by installed extensions.
    :rtype: collections.Iterator[types.ModuleType]

    """
    # Get the cached extension manager.
    try:
        m = _get_extension_plugin_modules.ext_manager
    except AttributeError:
        log.debug("Creating and caching ExtensionManager for namespace '{}'."
                  .format(EXTENSION_NAMESPACE))
        m = _get_extension_plugin_modules.ext_manager = \
            ExtensionManager(EXTENSION_NAMESPACE)
    # Yield registered extensions that are actually modules.
    for ext in m:
        ext_plugin_module = ext.plugin
        if not isinstance(ext_plugin_module, types.ModuleType):
            if warn:
                warnings.warn("Skipping extension provided by package '{} "
                              "({})' that did NOT resolve to a python module "
                              "(got an object of type {} instead: {})."
                              .format(ext.entry_point.dist.key,
                                      ext.entry_point.dist.version,
                                      type(ext_plugin_module).__name__,
                                      ext_plugin_module))
        else:
            yield ext.plugin


def get_plugins(interface_type, env_var, helper_var,
                warn=True, reload_modules=False):
    """
    Discover and return classes implementing the given ``interface_class``.

    Discoverable implementations may either be located in sub-modules parallel
    to the definition of the interface class or be located in modules specified
    in the environment variable  ``env_var``.

    In order to specify additional out-of-scope python modules containing
    interface-class implementations, additions to the given environment variable
    must be made. Entries must be separated by the standard PATH separating
    character based on the operating OS standard (e.g. ';' (for windows) or
    ':' for most everything else). Entries should be importable python module
    paths.

    When looking at module attributes, we only acknowledge those that start with
    an alphanumeric character. '_' prefixed attributes are effectively hidden
    from discovery by this function when merely scanning a module's attributes.

    We required that the base class that we are checking for also descends from
    the ``Pluggable`` interface defined above. This allows us to check if a
    loaded class ``is_usable``.

    Within a module we first look for a helper variable by the name provided,
    which can either be a single class object or an iterable of class objects,
    to be specifically exported. If the variable is set to None, we skip that
    module and do not import anything. If the variable is not present, we look
    at attributes defined in that module for classes that descend from the given
    base class type. If none of the above are found, or if an exception occurs,
    the module is skipped.

    :param type interface_type:
        Interface class type of which we want to discover implementations of
        (the plugins).

    :param str env_var:
        String name of an environment variable defining additional python module
        paths, whose child modules are searched for implementing sub-classes of
        the specified base type.

    :param str helper_var:
        Name of the expected module helper attribute.

    :param bool warn:
        If we should warn about module import failures.

    :param bool reload_modules:
        Explicitly reload discovered modules from source instead of taking a
        potentially cached version of the module.

    :return: Set of discovered class types descending from type
        ``interface_type`` and ``smqtk.utils.plugin.Pluggable`` whose keys are
        the string names of the class types.
    :rtype: set[type[Pluggable]]

    """
    if not issubclass(interface_type, Pluggable):
        raise ValueError("Required base-class must descend from the Pluggable "
                         "interface!")

    log = logging.getLogger('.'.join([
        __name__, 'get_plugins[{}]'.format(interface_type.__name__)
    ]))

    modules_iter = \
        itertools.chain(
            _get_local_plugin_modules(log, interface_type, warn=warn),
            _get_envvar_plugin_module(log, env_var, warn=warn),
            _get_extension_plugin_modules(log, warn=warn)
        )

    log.debug("Getting plugins for interface '{}'"
              .format(interface_type.__name__))
    class_set = set()
    for _module in modules_iter:
        module_path = _module.__name__
        log.debug("Examining module: {}".format(module_path))
        if reload_modules:
            # Invoke reload in case the module changed between imports.
            # six should find the right thing.
            # noinspection PyCompatibility
            _module = reload_module(_module)
            if _module is None:
                raise RuntimeError("[{}] Failed to reload".format(module_path))

        # Find valid classes in the discovered module by:
        classes = []
        if hasattr(_module, helper_var):
            # Looking for magic variable for import guidance
            classes = getattr(_module, helper_var)
            if classes is None:
                log.debug("[%s] Helper is None-valued, skipping module",
                          module_path)
                classes = []
            elif (isinstance(classes, collections.Iterable) and
                  not isinstance(classes, six.string_types)):
                classes = list(classes)
                log.debug("[%s] Loaded list of %d class types via helper",
                          module_path, len(classes))
            # Thus, non-iterable value.
            elif isinstance(classes, type) \
                    and issubclass(classes, interface_type):
                log.debug("[%s] Loaded class type: %s",
                          module_path, classes.__name__)
                classes = [classes]
            else:
                raise RuntimeError("[%s] Helper variable set to an invalid "
                                   "value: %s" % (module_path, classes))
        else:
            # Scan module valid attributes for classes that descend from the
            # given base-class.
            log.debug("[%s] No helper, scanning module attributes",
                      module_path)
            for attr_name in dir(_module):
                if VALUE_ATTRIBUTE_RE.match(attr_name):
                    classes.append(getattr(_module, attr_name))

        # Check the validity of the discovered class types in this module.
        for cls in classes:
            # check that all class types in iterable are:
            # - Class types,
            # - Subclasses of the given base-type and plugin interface
            # - Not missing any abstract implementations.
            #
            # noinspection PyUnresolvedReferences
            if not isinstance(cls, type):
                # No logging, over verbose, undetermined type.
                pass
            elif cls is interface_type:
                log.debug("[%s.%s] [skip] Literally the base class.",
                          module_path, cls.__name__)
            elif not issubclass(cls, interface_type):
                log.debug("[%s.%s] [skip] Does not descend from base class.",
                          module_path, cls.__name__)
            elif bool(cls.__abstractmethods__):
                # Making this a warning as I think this indicates a broken
                # implementation in the ecosystem.
                # noinspection PyUnresolvedReferences
                log.warn('[%s.%s] [skip] Does not implement one or more '
                         'abstract methods: %s',
                         module_path, cls.__name__,
                         list(cls.__abstractmethods__))
            elif not cls.is_usable():
                log.debug("[%s.%s] [skip] Class does not report as usable.",
                          module_path, cls.__name__)
            else:
                log.debug('[%s.%s] [KEEP] Retaining subclass.',
                          module_path, cls.__name__)
                class_set.add(cls)

    return class_set


class NotUsableError (Exception):
    """
    Exception thrown when a pluggable class is constructed but does not report
    as usable.
    """


@six.add_metaclass(abc.ABCMeta)
class Pluggable (object):
    """
    Interface for classes that have plugin implementations
    """

    __slots__ = ()

    PLUGIN_ENV_VAR = "SMQTK_PLUGIN_PATH"
    PLUGIN_HELPER_VAR = "SMQTK_PLUGIN_CLASS"

    @classmethod
    def get_impls(cls, warn=True, reload_modules=False):
        """
        Discover and return a set of classes that implement the calling class.

        See the ``get_plugins`` function for more details on the logic of how
        implementing classes (aka "plugins") are discovered.

        The class-level variables ``PLUGIN_ENV_VAR`` and ``PLUGIN_HELPER_VAR``
        may be overridden to change what environment and helper variable are
        looked for, respectively.

        :param bool warn:
        If we should warn about module import failures.

        :param bool reload_modules:
            Explicitly reload discovered modules from source.

        :return: Set of discovered class types descending from type
            ``interface_type`` and ``smqtk.utils.plugin.Pluggable`` whose keys
            are the string names of the class types.
        :rtype: set[type[Pluggable]]

        """
        # TODO: If reload is False, cache result or use cache
        # TODO: If True, re-cache new result.
        return get_plugins(cls, cls.PLUGIN_ENV_VAR, cls.PLUGIN_HELPER_VAR,
                           warn=warn, reload_modules=reload_modules)

    @classmethod
    @abc.abstractmethod
    def is_usable(cls):
        """
        Check whether this class is available for use.

        Since certain plugin implementations may require additional dependencies
        that may not yet be available on the system, this method should check
        for those dependencies and return a boolean saying if the implementation
        is usable.

        NOTES:
            - This should be a class method
            - When an implementation is deemed not usable, this should emit a
                warning detailing why the implementation is not available for
                use.

        :return: Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        raise NotImplementedError("is_usable class-method not implemented for "
                                  "class '%s'" % cls.__name__)

    def __init__(self):
        if not self.is_usable():
            raise NotUsableError("Implementation class '%s' is not currently "
                                 "usable." % self.__class__.__name__)



"""
Utility for transforming an input image in various standardized ways, saving
out those transformed images with standard namings. Transformations used are
configurable via a configuration file (JSON).

Configuration details:
{
    "crop": {

        "center_levels": null | int
            # If greater than 0, crop out one or more increasing smaller images
            # from a base image by cutting off increasingly larger portions of
            # the outside perimeter. Cropped image dimensions determined by the
            # dimensions of the base image and the number of crops to generate.

        "quadrant_pyramid_levels": null | int
            # If greater than 0, generate a number of crops based on a number of
            # quad-tree partitions made based on the given number of levels.
            # Partitions for all levels less than the level provides are also
            # made.

        "tile_shape": null | [width, height]
            # If not null and is a list of two integers, crop out tile windows
            # from the base image that have the width and height specified.
            # If the image width or height is not evenly divisible by the tile
            # width or height, respectively, then the crop out as many tiles as
            # neatly fit starting from the axis origin. The remaining pixels are
            # ignored.

        "tile_stride": null | [x, y]
            # If not null and is a list of two integers, crop out sub-images of
            # the above width and height (if given) with this stride. When not
            # this is not provided, the default stride is the same as the tile
            # width and height.
    },

    "brightness_levels": null | int
        # Generate a number of images with different brightness levels using
        # linear interpolation to choose levels between 0 (black) and 1
        # (original image) as well as between 1 and 2.
        # Results will not include contrast level 0, 1 or 2 images.

    "contrast_levels": null | int
        # Generate a number of images with different contrast levels using
        # linear interpolation to choose levels between 0 (black) and 1
        # (original image) as well as between 1 and 2.
        # Results will not include contrast level 0, 1 or 2 images.

}

"""

import logging
import os

import PIL.Image

import smqtk.utils.cli
import smqtk.utils.file
import smqtk.utils.parallel
from smqtk.utils.image import (
    image_crop_center_levels, image_crop_quadrant_pyramid, image_crop_tiles,
    image_brightness_intervals, image_contrast_intervals
)


def generate_image_transformations(image_path,
                                   crop_center_n, crop_quadrant_levels,
                                   crop_tile_shape, crop_tile_stride,
                                   brightness_intervals,
                                   contrast_intervals,
                                   output_dir=None,
                                   output_ext='.png'):
    """
    Transform an input image into different crops or other transforms,
    outputting results to the given output directory without overwriting or
    otherwise changing the input image.

    By default, if not told otherwise, we will write output images in the same
    directory as the source image. Output images share a core filename as the
    source image, but with extra suffix syntax to differentiate produced images
    from the original. Output images will share the same image extension as the
    source image.
    """
    log = logging.getLogger(__name__)

    abs_path = os.path.abspath(image_path)
    output_dir = output_dir or os.path.dirname(abs_path)
    smqtk.utils.file.safe_create_dir(output_dir)
    p_base = os.path.splitext(os.path.basename(abs_path))[0]
    p_ext = output_ext
    p_base = os.path.join(output_dir, p_base)
    image = PIL.Image.open(image_path).convert('RGB')

    def save_image(img, suffixes):
        """
        Save an image based on source image basename and an iterable of suffix
        parts that will be separated by periods.
        """
        fn = '.'.join([p_base] + list(suffixes)) + p_ext
        log.debug("Saving: %s", fn)
        img.save(fn)

    if crop_center_n:
        log.info("Computing center crops")
        tag = "crop_centers"
        for l, c in image_crop_center_levels(image, crop_center_n):
            save_image(c, [tag, str(l)])

    if crop_quadrant_levels:
        log.info("Computing quadrant crops")
        tag = "crop_quadrants"
        for l, (i, j), c in image_crop_quadrant_pyramid(image,
                                                        crop_quadrant_levels):
            save_image(c, [tag, str(l), "q_{:d}_{:d}".format(i, j)])

    if crop_tile_shape and crop_tile_shape[0] > 0 and crop_tile_shape[1] > 0:
        tag = "crop_tiles"
        t_width = crop_tile_shape[0]
        t_height = crop_tile_shape[1]
        log.info("Cropping %dx%d pixel tiles from images with stride %s"
                 % (t_width, t_height, crop_tile_stride))
        # List needed to iterate generator.
        list(smqtk.utils.parallel.parallel_map(
            lambda x, y, ii:
                save_image(ii, [tag,
                                '%dx%d+%d+%d' % (t_width, t_height, x, y)]),
            image_crop_tiles(image, t_width, t_height, crop_tile_stride)
        ))

    if brightness_intervals:
        log.info("Computing brightness variants")
        for b, i in image_brightness_intervals(image, brightness_intervals):
            save_image(i, ['brightness', str(b)])

    if contrast_intervals:
        log.info("Computing contrast variants")
        for c, i in image_contrast_intervals(image, contrast_intervals):
            save_image(i, ['contrast', str(c)])


def default_config():
    return {
        "crop": {
            # 0 means disabled
            "center_levels": None,
            # 0 means disabled, 2 meaning 2x2 and 4x4
            "quadrant_pyramid_levels": None,
            # Tile shape or None for no tiling
            "tile_shape": None,
            # The stride of tiles top crop out. This defaults to the height and
            # width of the tiles to create non-overlapping chips.
            "tile_stride": None,
        },
        "brightness_levels": None,
        "contrast_levels": None,
    }


def cli_parser():
    parser = smqtk.utils.cli.basic_cli_parser(__doc__)

    g_io = parser.add_argument_group("Input/Output")
    g_io.add_argument("-i", "--image",
                      help="Image to produce transformations for.")
    g_io.add_argument("-o", "--output",
                      help="Directory to output generated images to. By "
                           "default, if not told otherwise, we will write "
                           "output images in the same directory as the source "
                           "image. Output images share a core filename as the "
                           "source image, but with extra suffix syntax to "
                           "differentiate produced images from the original. "
                           "Output images will share the same image extension "
                           "as the source image.")
    return parser


def main():
    args = cli_parser().parse_args()
    config = smqtk.utils.cli.utility_main_helper(default_config, args)
    input_image_path = args.image
    output_dir = args.output

    if input_image_path is None:
        raise ValueError("No input image path given")

    crop_center_levels = config['crop']['center_levels']
    crop_quad_levels = config['crop']['quadrant_pyramid_levels']
    crop_tile_shape = config['crop']['tile_shape']
    crop_tile_stride = config['crop']['tile_stride']
    b_levels = config['brightness_levels']
    c_levels = config['contrast_levels']

    generate_image_transformations(
        input_image_path,
        crop_center_levels, crop_quad_levels,
        crop_tile_shape, crop_tile_stride,
        b_levels, c_levels,
        output_dir
    )


if __name__ == '__main__':
    main()



"""
Default values and instances for the ObjectDetection interface.
"""
# - Using the same default factory for ObjectDetector as the Classifier
#   interface.
# - Providing classifier default here for convenience.
# noinspection PyProtectedMember
from smqtk.algorithms.classifier._defaults import DFLT_CLASSIFIER_FACTORY  # lgtm[py/unused-import]
from smqtk.representation import DetectionElementFactory
from smqtk.representation.detection_element.memory \
    import MemoryDetectionElement

DFLT_DETECTION_FACTORY = DetectionElementFactory(
    MemoryDetectionElement, {}
)



"""
LICENCE
-------
Copyright 2013-2016 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.

"""
###
# In specific ordering for dependency resolution
#
from .base_object import SmqtkObject  # No internal util dependencies
from .cli import initialize_logging
from .content_type_validator import ContentTypeValidator
from .database_info import DatabaseInfo
from .iter_validation import check_empty_iterable
from .read_write_lock import ReaderUpdateException, DummyRWLock, ReadWriteLock
from .safe_config_comment_parser import SafeConfigCommentParser
from .signal_handler import SignalHandler
from .simple_timer import SimpleTimer



from smqtk.utils import SmqtkObject
from smqtk.utils.configuration import Configurable
from smqtk.utils.plugin import Pluggable


# noinspection PyAbstractClass
class SmqtkAlgorithm (SmqtkObject, Configurable, Pluggable):
    """
    Parent class for all algorithm interfaces.
    """

    __slots__ = ()

    @property
    def name(self):
        """
        :return: The name of this class type.
        :rtype: str
        """
        return self.__class__.__name__



import abc
import numpy

from collections import defaultdict

from smqtk.representation import SmqtkRepresentation
from smqtk.utils.dict import merge_dict
from smqtk.utils.plugin import Pluggable
from smqtk.utils.parallel import parallel_map

from ._io import elements_to_matrix


def _uuid_and_vector_from_descriptor(descriptor):
    """
    Given a descriptor, return a tuple containing the UUID and associated
    vector for that descriptor

    :param descriptor: The descriptor to process.
    :type descriptor: smqtk.representation.descriptor_element.DescriptorElement
    :return: Tuple containing the UUID and associated vector for the given
        descriptor
    :rtype: tuple[collections.Hashable, numpy.ndarray]
    """
    return (descriptor.uuid(), descriptor.vector())


class DescriptorElement (SmqtkRepresentation, Pluggable):
    """
    Abstract descriptor vector container.

    This structure supports implementations that cache descriptor vectors on a
    per-UUID basis.

    UUIDs must maintain unique-ness when transformed into a string.

    Descriptor element equality based on shared descriptor type and vector
    equality. Two descriptor vectors that are generated by different types of
    descriptor generator should not be considered the same (though, this may be
    up for discussion).

    Stored vectors should be effectively immutable.

    """

    def __init__(self, type_str, uuid):
        """
        Initialize a new descriptor element.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: Unique ID reference of the descriptor.
        :type uuid: collections.Hashable

        """
        super(DescriptorElement, self).__init__()

        self._type_label = type_str
        self._uuid = uuid

    def __hash__(self):
        return hash(self.uuid())

    def __eq__(self, other):
        if isinstance(other, DescriptorElement):
            return numpy.array_equal(self.vector(), other.vector())
        return False

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s{type: %s, uuid: %s}" % (self.__class__.__name__, self.type(),
                                           self.uuid())

    def __getstate__(self):
        return {
            "_type_label": self._type_label,
            "_uuid": self._uuid,
        }

    def __setstate__(self, state):
        self._type_label = state['_type_label']
        self._uuid = state['_uuid']

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        aside from the first two assumed positional arguments, turning those
        argument names into configuration dictionary keys.
        If any of those arguments have defaults, we will add those values into
        the configuration dictionary appropriately.
        The dictionary returned should only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        # similar to parent impl, except we remove the ``type_str`` and ``uuid``
        # configuration parameters as they are to be specified at runtime.
        dc = super(DescriptorElement, cls).get_default_config()
        # These parameters must be specified at construction time.
        del dc['type_str'], dc['uuid']
        return dc

    # noinspection PyMethodOverriding
    @classmethod
    def from_config(cls, config_dict, type_str, uuid, merge_default=True):
        """
        Instantiate a new instance of this class given the desired type, uuid,
        and JSON-compliant configuration dictionary.

        :param type_str: Type of descriptor. This is usually the name of the
            content descriptor that generated this vector.
        :type type_str: str

        :param uuid: Unique ID reference of the descriptor.
        :type uuid: collections.Hashable

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: DescriptorElement

        """
        c = {}
        merge_dict(c, config_dict)
        c['type_str'] = type_str
        c['uuid'] = uuid
        return super(DescriptorElement, cls).from_config(c, merge_default)

    def uuid(self):
        """
        :return: Unique ID for this vector.
        :rtype: collections.Hashable
        """
        return self._uuid

    def type(self):
        """
        :return: Type label type of the DescriptorGenerator that generated this
            vector.
        :rtype: str
        """
        return self._type_label

    @classmethod
    def _get_many_vectors(cls, descriptors):
        """
        Internal method to be overridden by subclasses to return many vectors
        associated with given descriptors.

        :note: Returned vectors are *not* guaranteed to be returned in the
            order they are requested. Missing vectors may be returned as None
            or omitted entirely from results. The wrapper function
            `get_many_vectors` handles re-ordering as necessary and insertion
            of None for missing values.

        :param descriptors: Iterable of descriptors to query for.
        :type descriptors: collections.Iterable[
            smqtk.representation.descriptor_element.DescriptorElement]

        :return: Iterator of tuples containing the descriptor uuid and the
            vector associated with the given descriptors or None if the
            descriptor has no associated vector
        :rtype: collections.Iterable[
            tuple[collections.Hashable, Union[numpy.ndarray, None]]]
        """
        for uuid_vector_pair in parallel_map(
                _uuid_and_vector_from_descriptor, descriptors,
                name='retrieve_vectors'):
            yield uuid_vector_pair

    @classmethod
    def get_many_vectors(cls, descriptors):
        """
        Get an iterator over vectors associated with given descriptors.

        :note: Most subclasses should override internal method
            `_get_many_vectors` rather than this external wrapper function. If
            a subclass does override this classmethod, it is responsible for
            appropriately handling any valid DescriptorElement, regardless of
            subclass.

        :param descriptors: Iterable of descriptors to query for.
        :type descriptors: collections.Iterable[
            smqtk.representation.descriptor_element.DescriptorElement]

        :return: Iterable of vectors associated with the given descriptors or
            None if the descriptor has no associated vector. Results are
            returned in the order that descriptors were given.
        :rtype: list[numpy.ndarray | None]
        """
        batch_dictionary = defaultdict(list)
        uuid_indices = {}
        index = -1
        for index, descriptor_ in enumerate(descriptors):
            # Divide descriptors up into batches based on their type, since
            # each DescriptorElement subclass knows best how to optimally
            # retrieve vectors of its own type.
            batch_dictionary[type(descriptor_)].append(descriptor_)
            # Keep track of the order of descriptors to ensure that we return
            # vectors in the requested order after batching them out.
            uuid_indices[descriptor_.uuid()] = index

        # Default to None, since _get_many_vectors implementations can ignore
        # any descriptors that cannot be retrieved
        ordered_vectors = [None] * (index + 1)

        # Retrieve all the vectors for a given type of descriptor in a single
        # batch
        for _cls, descriptor_batch in batch_dictionary.items():
            # noinspection PyProtectedMember
            for uuid, vector in _cls._get_many_vectors(descriptor_batch):
                ordered_vectors[uuid_indices[uuid]] = vector

        return ordered_vectors

    ###
    # Abstract methods
    #

    @abc.abstractmethod
    def has_vector(self):
        """
        :return: Whether or not this container current has a descriptor vector
            stored.
        :rtype: bool
        """

    @abc.abstractmethod
    def vector(self):
        """
        :return: Get the stored descriptor vector as a numpy array. This returns
            None of there is no vector stored in this container.
        :rtype: numpy.ndarray or None
        """

    @abc.abstractmethod
    def set_vector(self, new_vec):
        """
        Set the contained vector.

        If this container already stores a descriptor vector, this will
        overwrite it.

        :param new_vec: New vector to contain.
        :type new_vec: numpy.ndarray

        :returns: Self.
        :rtype: DescriptorMemoryElement

        """



import warnings

from smqtk.representation import DataSet
from smqtk.representation.data_element.psql import PostgresDataElement
from smqtk.utils.postgres import PsqlConnectionHelper

# Try to import required modules
try:
    import psycopg2
except ImportError:
    psycopg2 = None


class PostgresNativeDataSet (DataSet):
    """
    Dataset that stores data elements natively in a PostgreSQL database.

    Elements stored in this data set implementation will be copied as
    PostgresDataElements, which are stored based on this data set's
    configuration.

    Data elements retrieved from this data set will be of the
    PostgresDataElement class type.
    """

    @classmethod
    def is_usable(cls):
        if psycopg2 is None:
            warnings.warn("PostgresNativeDataSet not usable due to psycopg2 "
                          "package not being importable.")
            return False
        return True

    def __init__(self, table_name="psql_data_elements", id_col="id",
                 sha1_col="sha1", mime_col="mime", byte_col="bytes",
                 db_name="postgres", db_host="/tmp", db_port=5432,
                 db_user=None, db_pass=None, itersize=1000, read_only=False,
                 create_table=True):
        """
        Create a PostgreSQL-based data set instance.

        If the tabled mapped to the provided ``table_name`` already exists, we
        expect the provided columns to match the following types:
        - ``id_col`` is expected to be TEXT
        - ``sha1_col`` is expected to be TEXT
        - ``type_col`` is expected to be TEXT
        - ``byte_col`` is expected to be BYTEA

        Default database connection parameters are assuming the use of a
        non-default, non-postgres-user cluster where the current user's name is
        equivalent to a valid role in the database.

        :param str table_name:
            String label of the table in the database to interact with.
        :param str id_col:
            Name of the element ID column in ``table_name``.
        :param str sha1_col:
            Name of the SHA1 column in ``table_name``.
        :param str mime_col:
            Name of the MIMETYPE column in ``table_name``.
        :param str byte_col:
            Name of the column storing byte data in ``table_name``.
        :param str|None db_host:
            Host address of the PostgreSQL server. If None, we assume the
            server is on the local machine and use the UNIX socket. This might
            be a required field on Windows machines (not tested yet).
        :param str|None db_port:
            Port the Postgres server is exposed on. If None, we assume a
            default port (5433).
        :param str db_name:
            The name of the database to connect to.
        :param str|None db_user:
            Postgres user to connect as. If None, postgres defaults to using
            the current accessing user account name on the operating system.
        :param str|None db_pass:
            Password for the user we're connecting as. This may be None if no
            password is to be used.
        :param int itersize:
            Number of records fetched per network round trip when iterating
            over a named cursor.
        :param bool read_only:
            Only allow reading of this data.  Modification actions will throw a
            ReadOnlyError exceptions.
        :param bool create_table:
            If this instance should try to create the storing table before
            actions are performed against it. If the configured user does not
            have sufficient permissions to create the table and it does not
            currently exist, an exception will be raised.

        """
        super(PostgresNativeDataSet, self).__init__()

        itersize = int(itersize)
        if itersize <= 0:
            raise ValueError("Itersize must be greater than 0.")

        self._table_name = table_name
        self._id_col = id_col
        self._sha1_col = sha1_col
        self._mime_col = mime_col
        self._byte_col = byte_col
        self._read_only = read_only
        self._create_table = create_table

        self._psql_helper = PsqlConnectionHelper(db_name, db_host, db_port,
                                                 db_user, db_pass, itersize)

        # Set table creation SQL in helper
        if not self._read_only:
            self._psql_helper.set_table_upsert_sql(
                PostgresDataElement.CommandTemplates.UPSERT_TABLE.format(
                    table_name=self._table_name,
                    id_col=self._id_col,
                    sha1_col=self._sha1_col,
                    mime_col=self._mime_col,
                    byte_col=byte_col,
                )
            )

    def get_config(self):
        return {
            "table_name": self._table_name,
            "id_col": self._id_col,
            "sha1_col": self._sha1_col,
            "mime_col": self._mime_col,
            "byte_col": self._byte_col,

            "db_name": self._psql_helper.db_name,
            "db_host": self._psql_helper.db_host,
            "db_port": self._psql_helper.db_port,
            "db_user": self._psql_helper.db_user,
            "db_pass": self._psql_helper.db_pass,
            "itersize": self._psql_helper.itersize,

            "read_only": self._read_only,
            "create_table": self._create_table,
        }

    def _gen_psql_element(self, uid, content_type=None):
        """
        Internal method to generate a psql data element with appropriate psql
        parameters.
        :param collections.Hashable uid: UUID of data element.
        :param None|str content_type: Content type / MIME type of the element.
        """
        e = PostgresDataElement(
            uid, content_type=content_type, table_name=self._table_name,
            id_col=self._id_col, sha1_col=self._sha1_col,
            mime_col=self._mime_col, byte_col=self._byte_col,
            read_only=self._read_only, create_table=self._create_table
        )
        # Share PSQL helper instance.
        e._psql_helper = self._psql_helper
        return e

    def __iter__(self):
        """
        :return: Generator over the DataElements contained in this set in no
            particular order.
        """
        # Select all UUIDs and content type, yielding constructed psql
        # data elements.
        q = "SELECT {id_col:s}, {mime_col:s} FROM {table_name:s};".format(
            id_col=self._id_col,
            mime_col=self._mime_col,
            table_name=self._table_name,
        )

        def cb(cursor):
            cursor.execute(q)

        for r in self._psql_helper.single_execute(cb, yield_result_rows=True):
            e_uuid, e_mimetype = r
            yield self._gen_psql_element(e_uuid, e_mimetype)

    def count(self):
        """
        :return: The number of data elements in this set.
        :rtype: int
        """
        # Query count of rows in table (select on id col only)
        count_query = "SELECT count({id_col:s}) FROM {table_name:s};".format(
            id_col=self._id_col,
            table_name=self._table_name,
        )

        def cb(cursor):
            cursor.execute(count_query)

        r = list(self._psql_helper.single_execute(cb, yield_result_rows=True))
        if not r:
            # No rows in table
            return 0
        return int(r[0][0])

    def uuids(self):
        """
        :return: A new set of uuids represented in this data set.
        :rtype: set
        """
        # TODO: UPDATE TO ITERATOR INSTEAD OF SET RETURN TYPE
        q = "SELECT {id_col:s} FROM {table_name:s};".format(
            id_col=self._id_col,
            table_name=self._table_name,
        )

        def cb(cursor):
            cursor.execute(q)

        return {r[0] for r in
                self._psql_helper.single_execute(cb, yield_result_rows=True)}

    def has_uuid(self, uuid):
        """
        Test if the given uuid refers to an element in this data set.

        :param uuid: Unique ID to test for inclusion. This should match the
            type that the set implementation expects or cares about.
        :type uuid: collections.Hashable

        :return: True if the given uuid matches an element in this set, or
            False if it does not.
        :rtype: bool

        """
        # Query for table id col values
        q = "SELECT {id_col:s} FROM {table_name:s} "\
            "WHERE {id_col:s} = %(id_val)s;"\
            .format(id_col=self._id_col,
                    table_name=self._table_name)

        def cb(cursor):
            cursor.execute(q, {'id_val': str(uuid)})

        return bool(list(
            self._psql_helper.single_execute(cb, yield_result_rows=True)
        ))

    def add_data(self, *elems):
        """
        Add the given data element(s) instance to this data set.

        *NOTE: Implementing methods should check that input elements are in
        fact DataElement instances.*

        :param elems: Data element(s) to add
        :type elems: smqtk.representation.DataElement

        """
        # TODO: Optimize for batch insertion using custom query.
        for e in elems:
            pe = self._gen_psql_element(e.uuid(), e.content_type())
            pe.set_bytes(e.get_bytes())

    def get_data(self, uuid):
        """
        Get the data element the given uuid references, or raise an
        exception if the uuid does not reference any element in this set.

        :raises KeyError: If the given uuid does not refer to an element in
            this data set.

        :param uuid: The uuid of the element to retrieve.
        :type uuid: collections.Hashable

        :return: The data element instance for the given uuid.
        :rtype: smqtk.representation.DataElement

        """
        # Query for content type recorded in our table to use for PSQL element
        # construction.
        q = "SELECT {ct_col:s} FROM {table_name:s} " \
            "WHERE {id_col:s} = %(id_val)s;" \
            .format(ct_col=self._mime_col,
                    table_name=self._table_name,
                    id_col=self._id_col)

        def cb(cursor):
            cursor.execute(q, {'id_val': str(uuid)})

        r = list(self._psql_helper.single_execute(cb, yield_result_rows=True))
        if not r:
            # No rows matching the input uuid were found.
            raise KeyError(uuid)

        # Create and return the PSQL element.
        ct = str(r[0][0])
        return self._gen_psql_element(uuid, content_type=ct)



import numpy as np
from sklearn.linear_model import LogisticRegression

from smqtk.algorithms import SupervisedClassifier
from smqtk.representation import DescriptorElement


class SkLearnLogisticRegression (LogisticRegression, SupervisedClassifier):
    """
    Classifier implementation using Scikit Learn's LogisticRegression
    classifier.

    See ``sklearn.linear_model.LogisticRegression`` documentation for more
    details.
    """

    @classmethod
    def is_usable(cls):
        # scikit-learn is a dependency of SMQTK
        return True

    def get_config(self):
        return self.get_params()

    def has_model(self):
        try:
            return self.coef_ is not None
        except AttributeError:
            return False

    def get_labels(self):
        return self.classes_.tolist()

    def _train(self, class_examples, **extra_params):
        # convert descriptor elements into combines ndarray with associated
        # label vector.
        vec_list = []
        label_list = []
        for label, examples in class_examples.items():
            label_vectors = \
                DescriptorElement.get_many_vectors(examples)
            # ``is`` or ``count`` method messes up when elements are np arrays.
            none_count = len([e for e in label_vectors if e is None])
            assert none_count == 0, \
                "Some descriptor elements for label {} did not contain " \
                "vectors! (n={})".format(label, none_count)
            vec_list.extend(label_vectors)
            label_list.extend([label] * len(label_vectors))
        vec_list = np.vstack(vec_list)
        self.fit(vec_list, label_list)

    def _classify_arrays(self, array_iter):
        # Collect arrays for prediction.
        # - Collect into numpy.ndarray if not already one.
        if isinstance(array_iter, np.ndarray):
            mat = array_iter
        else:
            # Expand out input iterable of arrays into a large matrix for
            # prediction.
            mat = np.array(list(array_iter))
        proba_mat = self.predict_proba(mat)
        class_list = self.classes_.tolist()
        for proba in proba_mat:
            yield dict(zip(class_list, proba))



"""
Utility functions pertaining to python dictionaries.
"""

import copy


def merge_dict(a, b, deep_copy=False):
    """
    Merge dictionary b into dictionary a.

    This is different than normal dictionary update in that we don't bash
    nested dictionaries, instead recursively updating them.

    For congruent keys, values are are overwritten, while new keys in ``b`` are
    simply added to ``a``.

    Values are assigned (not copied) by default. Setting ``deep_copy`` causes
    values from ``b`` to be deep-copied into ``a``.

    :param a: The "base" dictionary that is updated in place.
    :type a: dict

    :param b: The dictionary to merge into ``a`` recursively.
    :type b: dict

    :param deep_copy: Optionally deep-copy values from ``b`` when assigning into
        ``a``.
    :type deep_copy: bool

    :return: ``a`` dictionary after merger (not a copy).
    :rtype: dict

    """
    for k in b:
        if k in a and isinstance(a[k], dict) and isinstance(b[k], dict):
            merge_dict(a[k], b[k], deep_copy)
        elif deep_copy:
            a[k] = copy.deepcopy(b[k])
        else:
            a[k] = b[k]
    return a



import logging


class SmqtkObject (object):
    """
    Highest level object interface for classes defined in SMQTK.

    Currently defines logging methods.

    """

    __slots__ = ()

    @classmethod
    def get_logger(cls):
        """
        :return: logging object for this class
        :rtype: logging.Logger
        """
        return logging.getLogger('.'.join((cls.__module__, cls.__name__)))

    @property
    def _log(self):
        """
        :return: logging object for this class as a property
        :rtype: logging.Logger
        """
        return self.get_logger()



import numpy as np
from six.moves import zip

from smqtk.algorithms import SupervisedClassifier
from smqtk.representation import DescriptorElement
from smqtk.utils.configuration import (
    from_config_dict,
    make_default_config,
    cls_conf_to_config_dict,
)

from . import RelevancyIndex, NoIndexError


class SupervisedClassifierRelevancyIndex (RelevancyIndex):
    """
    Relevancy index that utilizes a usable supervised classifier for on-the-fly
    training and inference.

    # Classifier "cloning"
    The input supervised classifier instance to the constructor is not directly
    used, but its type and configuration are recorded in order to create a new
    instance in ``rank`` to train and classify the index.

    The caveat here is that any non-configuration reflected, runtime
    modifications to the input classifier will not be reflected by the
    classifier used in ``rank``.

    Using a copy of the input classifier allows the ``rank`` method to be used
    in parallel without blocking other calls to ``rank``.

    :param smqtk.algorithms.SupervisedClassifier classifier_inst:
        Supervised classifier instance to base the ephemeral ranking classifier
        on. The type and configuration of this classifier is used to create a
        clone at rank time. The input classifier instance is not modified.
    """

    def __init__(self, classifier_inst):
        super(SupervisedClassifierRelevancyIndex, self).__init__()
        self._classifier_type = type(classifier_inst)
        self._classifier_config = classifier_inst.get_config()
        # Some number of descriptors to be ranked, cached as the elements
        # themselves as well as a vertically-stacked matrix (ndim==2).
        # These are None when there is no index yet.
        self._descr_elem_list = None
        self._descr_matrix = None

    @classmethod
    def is_usable(cls):
        # This being a wrapper to other plugins, this is always available
        # and is more contingent on nested implementations existing.
        return True

    @classmethod
    def get_default_config(cls):
        c = super(SupervisedClassifierRelevancyIndex, cls).get_default_config()
        c['classifier_inst'] = \
            make_default_config(SupervisedClassifier.get_impls())
        return c

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        config_dict = dict(config_dict)  # shallow copy to write to input dict
        config_dict['classifier_inst'] = \
            from_config_dict(config_dict.get('classifier_inst', {}),
                             SupervisedClassifier.get_impls())
        return super(SupervisedClassifierRelevancyIndex, cls).from_config(
            config_dict, merge_default=merge_default
        )

    def get_config(self):
        return {
            'classifier_inst':
                cls_conf_to_config_dict(self._classifier_type,
                                        self._classifier_config),
        }

    def count(self):
        return 0 if self._descr_matrix is None else len(self._descr_matrix)

    def build_index(self, descriptors):
        # Cache given descriptor element vectors into a matrix for use during
        # ``rank``.
        descr_elem_list = list(descriptors)
        if len(descr_elem_list) == 0:
            raise ValueError("No descriptor elements passed.")
        # note: this fails if multiple descriptor elements with the same UID
        #       are included. There will be None's present.
        descr_matrix = np.asarray(
            DescriptorElement.get_many_vectors(descr_elem_list)
        )
        # If the result matrix is of dtype(object), then either some elements
        # did not have vectors or some vectors were not of congruent
        # dimensionality.
        if descr_matrix.dtype == np.dtype(object):
            raise ValueError("One or more descriptor elements did not have a "
                             "vector set or were of congruent dimensionality.")
        self._descr_elem_list = descr_elem_list
        self._descr_matrix = descr_matrix

    def rank(self, pos, neg):
        if self._descr_elem_list is None or self._descr_matrix is None:
            raise NoIndexError("No index built before calling rank.")

        # Train supervised classifier with positive/negative examples.
        label_pos = 'pos'
        label_neg = 'neg'

        classifier = self._classifier_type.from_config(self._classifier_config)
        classifier.train({
            label_pos: pos,
            label_neg: neg,
        })

        # Report ``label_pos`` class probabilities as rank score.
        cd_iter = zip(classifier.classify_arrays(self._descr_matrix),
                      self._descr_elem_list)
        report = {}
        for c_map, d_elem in cd_iter:
            report[d_elem] = c_map.get(label_pos, 0.0)
        return report



"""
Data element for storing bytes in PostgreSQL.

Note about PSQL and clusters
----------------------------
To create a cluster::

    $ sudo pg_createcluster -u $(whoami) 9.5 squadx

This does not start the cluster, which should be started with the following::

    $ pg_ctlcluster 9.5 squadx start

Note that we do not provide the ``--start`` flag to the pg_createcluster
command.  This because the postgresql server fails to start or stay active when
using the recommended systemctl-based commands.  Instead, pg_ctlcluster should
be used to start/stop the database.

"""
import hashlib
from threading import RLock

import six

from smqtk.representation import DataElement
from smqtk.utils.postgres import norm_psql_cmd_string, PsqlConnectionHelper

# Try to import required modules
try:
    import psycopg2
except ImportError:
    psycopg2 = None


# Lock for data element create-table functionality
GLOBAL_PSQL_TABLE_CREATE_RLOCK = RLock()


class PostgresDataElement (DataElement):  # lgtm [py/missing-equals]
    """
    Data element bytes stored in PostgreSQL database.

    Storage table should have three columns for the following components:
    - data SHA1 (effective UID)
    - data content-type / MIMETYPE
    - data bytes

    Efficient connection pooling may be achieved via external utilities like
    PGBounder.

    Due to the use of the "ON CONFLICT" clause in upserting data, this
    implementation requires at least PostgreSQL version 9.5 or greater.

    """

    # SHA1 checksum of 0-length data (empty bytes)
    EMPTY_SHA = hashlib.sha1(b'').hexdigest()

    class CommandTemplates (object):
        """ Encapsulation of command templates. """

        # Upsert table for storage if desired
        #
        # Format params:
        # - table_name
        # - id_col
        # - sha1_col
        # - mime_col
        # - byte_col
        UPSERT_TABLE = norm_psql_cmd_string("""
            CREATE TABLE IF NOT EXISTS {table_name:s} (
              {id_col:s}   TEXT NOT NULL,
              {sha1_col:s} TEXT NOT NULL,
              {mime_col:s} TEXT NOT NULL,
              {byte_col:s} BYTEA NOT NULL,
              PRIMARY KEY ({id_col:s})
            );
        """)

        # Select ``col`` for a given entry ID.
        #
        # Query Format params:
        # - col
        # - table_name
        # - id_col
        #
        # Value params:
        # - id_val
        SELECT = norm_psql_cmd_string("""
            SELECT {col:s}
              FROM {table_name:s}
              WHERE {id_col:s} = %(id_val)s
            ;
        """)

        # Upsert content-type/data for a uid
        #
        # Query Format params:
        # - table_name
        # - id_col
        # - sha1_col
        # - mime_col
        # - byte_col
        #
        # Value params:
        # - id_val
        # - sha1_val
        # - mime_val
        # - byte_val
        #
        # SQL format from:
        #   https://hashrocket.com/blog/posts/upsert-records-with-postgresql-9-5
        #
        UPSERT_DATA = norm_psql_cmd_string("""
            INSERT INTO {table_name:s} ({id_col:s}, {sha1_col:s}, {mime_col:s}, {byte_col:s})
                VALUES ( %(id_val)s, %(sha1_val)s, %(mime_val)s, %(byte_val)s )
                ON CONFLICT ({id_col:s})
                    DO UPDATE
                        SET ({sha1_col:s}, {mime_col:s}, {byte_col:s})
                          = (EXCLUDED.{sha1_col:s}, EXCLUDED.{mime_col:s}, EXCLUDED.{byte_col:s})
            ;
        """)

        # Same as ``UPSERT_DATA`` but does not set the mimetype on an update.
        # This is meant to atomically update the byte data without changing the
        # existing mimetype.
        UPSERT_DATA_NO_MIME = norm_psql_cmd_string("""
            INSERT INTO {table_name:s} ({id_col:s}, {sha1_col:s}, {mime_col:s}, {byte_col:s})
                VALUES ( %(id_val)s, %(sha1_val)s, %(mime_val)s, %(byte_val)s )
                ON CONFLICT ({id_col:s})
                    DO UPDATE
                        SET ({sha1_col:s}, {byte_col:s})
                          = (EXCLUDED.{sha1_col:s}, EXCLUDED.{byte_col:s})
            ;
        """)

    @classmethod
    def is_usable(cls):
        if psycopg2 is None:
            cls.get_logger().warning("Not usable. "
                                     "Requires the psycopg2 module.")
            return False
        return True

    def __init__(self, element_id, content_type=None,
                 table_name="psql_data_elements", id_col="id",
                 sha1_col="sha1", mime_col="mime", byte_col="bytes",
                 db_name="postgres", db_host="/tmp", db_port=5433, db_user=None,
                 db_pass=None, read_only=False,
                 create_table=True):
        """
        Create a new PostgreSQL-based data element.

        If the tabled mapped to the provided ``table_name`` already exists, we
        expect the provided columns to match the following types:
        - ``id_col`` is expected to be TEXT
        - ``sha1_col`` is expected to be TEXT
        - ``type_col`` is expected to be TEXT
        - ``byte_col`` is expected to be BYTEA

        Default database connection parameters are assuming the use of a
        non-default, non-postgres-user cluster where the current user's name is
        equivalent to a valid role in the database.

        :param element_id: ID to reference a specific data element row in the
            table.  This is required in the same way that a path is required to
            point to a file on a filesystem.
        :type element_id: str

        :param content_type: Expected mime-type of byte data set to this
            element.  This only affects setting the mime-type field when setting
            new bytes.  ``content_type()`` will always reflect what is stored in
            the backend, or lack there-of.

            If this mime-type differs from an existing stored value,
            this mime-type will overwrite the stored value on the next call to
            ``set_bytes``.  If this is None and there is no mime-type already
            set in the database, no mime-type will be set on the next
            ``set_bytes`` call.
        :type content_type: str | None

        :param table_name: String label of the table in the database to interact
            with.
        :type table_name: str

        :param id_col: Name of the element ID column in ``table_name``.
        :type id_col: str

        :param sha1_col: Name of the SHA1 column in ``table_name``.
        :type sha1_col: str

        :param mime_col: Name of the MIMETYPE column in ``table_name``.
        :type mime_col: str

        :param byte_col: Name of the column storing byte data in ``table_name``.
        :type byte_col: str

        :param db_host: Host address of the PostgreSQL server. If None, we
            assume the server is on the local machine and use the UNIX socket.
            This might be a required field on Windows machines (not tested yet).
        :type db_host: str | None

        :param db_port: Port the Postgres server is exposed on. If None, we
            assume a default port (5433).
        :type db_port: int | None

        :param db_name: The name of the database to connect to.
        :type db_name: str

        :param db_user: Postgres user to connect as. If None, postgres
            defaults to using the current accessing user account name on the
            operating system.
        :type db_user: str | None

        :param db_pass: Password for the user we're connecting as. This may be
            None if no password is to be used.
        :type db_pass: str | None

        :param read_only: Only allow reading of this data.  Modification actions
            will throw a ReadOnlyError exceptions.
        :type read_only: bool

        :param create_table: If this instance should try to create the storing
            table before actions are performed against it. If the configured
            user does not have sufficient permissions to create the table and it
            does not currently exist, an exception will be raised.
        :type create_table: bool

        """
        super(PostgresDataElement, self).__init__()

        if not isinstance(element_id, six.string_types):
            raise ValueError("Element ID should be a string type.")

        self._element_id = element_id
        self._content_type = content_type
        self._table_name = table_name

        self._id_col = id_col
        self._sha1_col = sha1_col
        self._mime_col = mime_col
        self._byte_col = byte_col

        self._read_only = read_only
        self._create_table = create_table

        # itersize is hard-coded because a single-element perspective should
        # only be retrieving one row at a time.
        self._psql_helper = PsqlConnectionHelper(db_name, db_host, db_port,
                                                 db_user, db_pass, 10,
                                                 GLOBAL_PSQL_TABLE_CREATE_RLOCK)

        # Set table creation SQL in helper
        if not self._read_only:
            self._psql_helper.set_table_upsert_sql(
                self.CommandTemplates.UPSERT_TABLE.format(
                    table_name=self._table_name,
                    id_col=self._id_col,
                    sha1_col=self._sha1_col,
                    mime_col=self._mime_col,
                    byte_col=byte_col,
                )
            )

    def __repr__(self):
        return "{:s}[id=\"{:s}\"]" \
            .format(self.__class__.__name__, self._element_id)

    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {
            "element_id": self._element_id,
            "table_name": self._table_name,
            "id_col": self._id_col,
            "sha1_col": self._sha1_col,
            "mime_col": self._mime_col,
            "byte_col": self._byte_col,

            "db_name": self._psql_helper.db_name,
            "db_host": self._psql_helper.db_host,
            "db_port": self._psql_helper.db_port,
            "db_user": self._psql_helper.db_user,
            "db_pass": self._psql_helper.db_pass,

            "read_only": self._read_only,
            "create_table": self._create_table,
        }

    def content_type(self):
        """
        :return: Standard type/subtype string for this data element, or None if
            the content type is unknown.
        :rtype: str or None
        """
        q = self.CommandTemplates.SELECT.format(
            col=self._mime_col,
            table_name=self._table_name,
            id_col=self._id_col,
        )
        v = dict(
            id_val=self._element_id
        )

        def cb(cursor):
            """
            :type cursor: psycopg2._psycopg.cursor
            """
            cursor.execute(q, v)

        r = list(self._psql_helper.single_execute(cb, yield_result_rows=True))
        if not r:
            return None
        elif len(r) > 1:
            raise RuntimeError("Somehow found multiple entries for the same"
                               "element ID (there should only be one).")
        return r[0][0]

    def is_empty(self):
        """
        Check if this element contains no bytes.

        The intent of this method is to quickly check if there is any data
        behind this element, ideally without having to read all/any of the
        underlying data.

        :return: If this element contains 0 bytes.
        :rtype: bool

        """
        q = self.CommandTemplates.SELECT.format(
            col="octet_length(%s)" % self._byte_col,
            table_name=self._table_name,
            id_col=self._id_col,
        )
        v = dict(
            id_val=self._element_id
        )

        def cb(cursor):
            """
            :type cursor: psycopg2._psycopg.cursor
            """
            cursor.execute(q, v)

        r = list(self._psql_helper.single_execute(cb, yield_result_rows=True))
        if not r:
            # No rows returned, meaning not entry for our element ID and no
            # bytes stored.
            return True
        elif len(r) > 1:
            raise RuntimeError("Somehow found multiple entries for the same"
                               "element ID (there should only be one).")

        num_bytes = int(r[0][0])
        if num_bytes == 0:
            # There was an entry, but the number of bytes stored was zero.
            return True
        else:
            # Non-zero number of bytes stored.
            return False

    def sha1(self):
        """
        Get the SHA1 checksum of this element's binary content.

        :return: SHA1 hex checksum of the data content.
        :rtype: str
        """
        q = self.CommandTemplates.SELECT.format(
            col=self._sha1_col,
            table_name=self._table_name,
            id_col=self._id_col,
        )
        v = dict(
            id_val=self._element_id,
        )

        def cb(cursor):
            """
            :type cursor: psycopg2._psycopg.cursor
            """
            cursor.execute(q, v)

        r = list(self._psql_helper.single_execute(cb, yield_result_rows=True))
        if not r:
            # no rows for element ID, so no bytes. Return SHA1 of empty string
            return self.EMPTY_SHA
        return r[0][0]

    def get_bytes(self):
        """
        :return: Get the bytes for this data element.
        :rtype: bytes
        """
        q = self.CommandTemplates.SELECT.format(
            col=self._byte_col,
            table_name=self._table_name,
            id_col=self._id_col,
        )
        v = dict(
            id_val=self._element_id
        )

        def cb(cursor):
            """
            :type cursor: psycopg2._psycopg.cursor
            """
            cursor.execute(q, v)

        r = list(self._psql_helper.single_execute(cb, yield_result_rows=True))
        if not r or len(r[0][0]) == 0:
            # No returned rows for element ID or if no bytes are stored.
            return bytes()
        else:
            return bytes(r[0][0])

    def writable(self):
        """
        :return: if this instance supports setting bytes.
        :rtype: bool
        """
        return not self._read_only

    def set_bytes(self, b):
        """
        Set bytes to this data element.

        Not all implementations may support setting bytes (check ``writable``
        method return).

        This base abstract method should be called by sub-class implementations
        first. We check for mutability based on ``writable()`` method return.

        :param b: bytes to set.
        :type b: byte

        :raises ReadOnlyError: This data element can only be read from / does
            not support writing.

        """
        super(PostgresDataElement, self).set_bytes(b)

        b_sha1 = hashlib.sha1(b).hexdigest()

        # TODO: Fallback to ``content_type()`` return if none provided in self.
        if self._content_type:
            # We have a content/mime type override as specified at element
            # construction.
            b_mimetype = self._content_type
            q_tmpl = self.CommandTemplates.UPSERT_DATA
        else:
            # Leave the mimetype alone or set an empty mimetype (none specified
            # at construction).
            b_mimetype = ""
            q_tmpl = self.CommandTemplates.UPSERT_DATA_NO_MIME

        q = q_tmpl.format(
            table_name=self._table_name,
            id_col=self._id_col,
            sha1_col=self._sha1_col,
            mime_col=self._mime_col,
            byte_col=self._byte_col,
        )
        v = dict(
            id_val=self._element_id,
            sha1_val=b_sha1,
            mime_val=b_mimetype,
            byte_val=psycopg2.Binary(b)
        )

        def cb(cursor):
            """
            :type cursor: psycopg2._psycopg.cursor
            """
            # TODO: Could be smart here and only update if content-type/byte
            #       data differs while keeping a row-lock between queries.
            cursor.execute(q, v)

        list(self._psql_helper.single_execute(cb))



import time

import flask
import requests

from smqtk.utils.dict import merge_dict


def make_response_json(message, return_code=200, **params):
    """
    Basic message constructor for returning JSON from a flask routing function

    :param message: String descriptive message to send back.
    :type message: str

    :param return_code: HTTP return code for this message. Default is 200.
    :type return_code: int

    :param params: Other key-value data to include in response JSON.
    :type params: JSON-compliant

    :return: Flask response and HTTP status code pair.
    :rtype: (flask.Response, int)

    """
    r = {
        "message": message,
        "time": {
            "unix": time.time(),
            "utc": time.asctime(time.gmtime()),
        }
    }
    merge_dict(r, params)
    return flask.jsonify(**r), return_code


class ServiceProxy (object):
    """
    Helper class for interacting with an external service.
    """

    def __init__(self, url):
        """
        Parameters
        ---
            url : str
                URL to base requests on.
        """
        # Append http:// to the head of the URL if neither http(s) are present
        if not (url.startswith('http://') or url.startswith('https://')):
            url = 'http://' + url
        self.url = url

    def _compose(self, endpoint):
        return '/'.join([self.url, endpoint])

    def get(self, endpoint, **params):
        # Make params None if its empty.
        params = params and params or None
        return requests.get(self._compose(endpoint), params)

    def post(self, endpoint, **params):
        # Make params None if its empty.
        params = params and params or None
        return requests.post(self._compose(endpoint), data=params)

    def put(self, endpoint, **params):
        # Make params None if its empty.
        params = params and params or None
        return requests.put(self._compose(endpoint), data=params)

    def delete(self, endpoint, **params):
        # Make params None if its empty.
        params = params and params or None
        return requests.delete(self._compose(endpoint), params=params)
